# distutils: language = c++
# distutils: extra_compile_args = -std=c++11

import numpy as np
cimport numpy as np
cimport cpython
#from cython.parallel import prange
DTYPE   = np.float
ctypedef np.float_t DTYPE_t
from numpy.math cimport INFINITY

import pyross.deterministic
import pyross.stochastic

from timeit import default_timer as timer
import time




cdef class SIR:
    """
    Susceptible, Infected, Removed (SIR)
    Ia: asymptomatic
    Is: symptomatic

    ...

    Parameters
    ----------
    parameters: dict
        Contains the following keys:
            alpha: float
                Estimate mean value of fraction of infected who are asymptomatic.
            beta: float
                Estimate mean value of rate of spread of infection.
            gIa: float
                Estimate mean value of rate of removal from asymptomatic individuals.
            gIs: float
                Estimate mean value of rate of removal from symptomatic individuals.
            fsa: float
                fraction by which symptomatic individuals do not self-isolate.
            cov: np.array( )
                covariance matrix for all the estimated parameters.
    M: int
        Number of compartments of individual for each class.
        I.e len(contactMatrix)
    Ni: np.array(3*M, )
        Initial number in each compartment and class

    Methods
    -------
    simulate
    """
    cdef:
        readonly int N, M,
        readonly double alpha, beta, gIa, gIs, fsa
        readonly np.ndarray rp0, Ni, drpdt, lld, CM, CC, means, cov

    def __init__(self, parameters, M, Ni):
        self.alpha = parameters.get('alpha')                    # fraction of asymptomatic infectives
        self.beta  = parameters.get('beta')                     # infection rate
        self.gIa   = parameters.get('gIa')                      # removal rate of Ia
        self.gIs   = parameters.get('gIs')                      # removal rate of Is
        self.fsa   = parameters.get('fsa')                      # the self-isolation parameter
        #
        self.cov = parameters.get('cov')
        #
        self.means = np.array([self.alpha,self.beta,self.gIa,self.gIs],dtype=DTYPE)
        self.cov = parameters.get('cov')

        self.N     = np.sum(Ni)
        self.M     = M
        self.Ni    = np.zeros( self.M, dtype=DTYPE)             # # people in each age-group
        self.Ni    = Ni


    def simulate(self, S0, Ia0, Is0, contactMatrix=None,
                Tf=100, Nf=101,
                Ns=1000,
                int nc=30, double epsilon = 0.03,
                int tau_update_frequency = 1,
                verbose=False,
                method='deterministic',
                events=[],contactMatrices=[],
                events_repeat=False,
                events_subsequent=True,
                ):
        """
        Parameters
        ----------
        S0: np.array(M,)
            Initial number of susceptables.
        Ia0: np.array(M,)
            Initial number of asymptomatic infectives.
        Is0: np.array(M,)
            Initial number of symptomatic infectives.
        contactMatrix: python function(t), optional
             The social contact matrix C_{ij} denotes the 
             average number of contacts made per day by an 
             individual in class i with an individual in class j
             The default is None.
        Tf: float, optional
            Final time of integrator. The default is 100.
        Nf: Int, optional
            Number of time points to evaluate.The default is 101,
        Ns: int, optional
            Number of samples of parameters to take. The default is 1000.
        nc: int, optional
        epsilon: np.float64, optional
            Acceptable error in leap. The default is 0.03.
        tau_update_frequency : int, optional
        verbose: bool, optional
            Verbosity of output. The default is False.
        Ti: float, optional
            Start time of integrator. The default is 0.
        method: str, optional
            Pyross integrator to use. The default is "deterministic".
        events: list of python functions, optional
            List of events that the current state can satisfy to change behaviour
            of the contact matrix. Event occurs when the value of the function
            changes sign. Event.direction determines which direction triggers
            the event, takign values {+1,-1}.
            The default is [].
        contactMatricies: list of python functions
            New contact matrix after the corresponding event occurs
            The default is [].
        events_repeat: bool, optional
            Wheither events is periodic in time. The default is false.
        events_subsequent : bool, optional
            TODO
             
        Returns
        -------
        out_dict : dict
            Dictionary containing the following keys:
                X: list
                    List of resultant trajectories
                t: list
                    List of times at which X is evaluated.
                X_mean : list
                    Mean trajectory of X
                X_std : list
                    Standard devation of trajectories of X at each time point.
                <init params> : Initial parameters passed at object instantiation.
                sample_parameters : list of parameters sampled to make trajectories.
        """
        cdef:
            int M=self.M
            double [:] means  = self.means
            double [:,:] cov  = self.cov
            double start_time, end_time
            np.ndarray trajectories = np.zeros( [Ns,3*M,Nf] , dtype=DTYPE)
            np.ndarray mean_traj = np.zeros([ 3*M, Nf] ,dtype=DTYPE)
            np.ndarray std_traj = np.zeros([ 3*M, Nf] ,dtype=DTYPE)
            np.ndarray sample_parameters
            bint control = False

        if (contactMatrix == None) and (len(events) == 0):
            raise RuntimeError("Neither contactMatrix function nor list of events provided.")

        if len(events) > 0:
            control = True

        sample_parameters = np.random.multivariate_normal(means, cov, Ns)

        start_time = timer()
        for i in range(Ns):
            if verbose:
                print('Running simulation {0} of {1}'.format(i+1,Ns),end='\r')
            while (sample_parameters[i] < 0).any():
                sample_parameters[i] = np.random.multivariate_normal(means,cov)
            parameters = {'fsa':self.fsa,
                        'alpha':sample_parameters[i,0],
                        'beta':sample_parameters[i,1],
                        'gIa':sample_parameters[i,2],'gIs':sample_parameters[i,3]}
            #
            if control:
                model = pyross.control.SIR(parameters, M, self.Ni)
                cur_result = model.simulate(S0, Ia0, Is0,
                              events, contactMatrices, Tf, Nf,
                              events_repeat=events_repeat,
                              events_subsequent=events_subsequent,
                              nc=nc,epsilon=epsilon,
                              tau_update_frequency=tau_update_frequency,
                              method=method)
            else:
                if method == 'deterministic':
                    model = pyross.deterministic.SIR(parameters, M, self.Ni)
                    cur_result = model.simulate(S0, Ia0, Is0,
                                    contactMatrix,
                                    Tf, Nf)
                else:
                    model = pyross.stochastic.SIR(parameters, M, self.Ni)
                    cur_result = model.simulate(S0, Ia0, Is0, contactMatrix, Tf, Nf,
                                  nc=nc,epsilon=epsilon,
                                  tau_update_frequency=tau_update_frequency,
                                  method=method)
            #
            trajectories[i] = cur_result['X'].T
        end_time = timer()
        if verbose:
            print('Finished. Time needed for evaluation:',time.strftime('%H:%M:%S',time.gmtime( end_time-start_time)) )
        #

        mean_traj = np.mean(trajectories,axis=0)
        std_traj = np.sqrt( np.var(trajectories,axis=0) )

        out_dict = {'X':trajectories, 't':cur_result['t'],
                    'X_mean':mean_traj,'X_std':std_traj,
                     'Ni':self.Ni, 'M':self.M,
                     'alpha':self.alpha, 'beta':self.beta,
                     'gIa':self.gIa, 'gIs':self.gIs,
                     'cov':self.cov,
                     'sample_parameters':sample_parameters
                      } #
        return out_dict




cdef class SIR_latent:
    """
    Susceptible, Infected, Removed (SIR)
    Ia: asymptomatic
    Is: symptomatic
    
    Latent inference class to be used when observed data is incomplete.
    ...

    Parameters
    ----------
    parameters: dict
        Contains the following keys:
            alpha: float
                Estimate mean value of fraction of infected who are asymptomatic.
            beta: float
                Estimate mean value of rate of spread of infection.
            gIa: float
                Estimate mean value of rate of removal from asymptomatic individuals.
            gIs: float
                Estimate mean value of rate of removal from symptomatic individuals.
            fsa: float
                fraction by which symptomatic individuals do not self-isolate.
            cov: np.array( )
                Covariance matrix for all the estimated parameters.
            S0: np.array(M,)
                Estimate initial number of susceptables.
            Ia0: np.array(M,)
                Estimate initial number of asymptomatic infectives.
            Is0: np.array(M,)
                Estimate initial number of symptomatic infectives.
            cov_init : np.array((3*M, 3*M)) :
                Covariance matrix for the initial state.
                
    M: int
        Number of compartments of individual for each class.
        I.e len(contactMatrix)
    Ni: np.array(3*M, )
        Initial number in each compartment and class

    Methods
    -------
    simulate
    """
    cdef:
        readonly int N, M, k_random
        readonly double alpha, beta, gIa, gIs, fsa
        readonly np.ndarray rp0, Ni, drpdt, lld, CM, CC, means_params, means_init, cov_params, cov_init

    def __init__(self, parameters, M, Ni):
        self.alpha = parameters.get('alpha')                    # fraction of asymptomatic infectives
        self.beta  = parameters.get('beta')                     # infection rate
        self.gIa   = parameters.get('gIa')                      # removal rate of Ia
        self.gIs   = parameters.get('gIs')                      # removal rate of Is
        self.fsa   = parameters.get('fsa')                      # the self-isolation parameter
        #
        self.cov_params = parameters.get('cov_params')
        self.cov_init = parameters.get('cov_init')
        #
        self.k_random = 4
        self.means_params = np.array([self.alpha,self.beta,self.gIa,self.gIs],dtype=DTYPE)
        self.means_init = np.zeros(3*M,dtype=DTYPE)
        self.means_init[: M] = np.array(parameters.get('S0'),dtype=DTYPE)
        self.means_init[1*M:2*M] = np.array(parameters.get('Ia0'),dtype=DTYPE)
        self.means_init[2*M:3*M] = np.array(parameters.get('Is0'),dtype=DTYPE)
        #
        self.N     = np.sum(Ni)
        self.M     = M
        self.Ni    = np.zeros( self.M, dtype=DTYPE)             # # people in each age-group
        self.Ni    = Ni


    def simulate(self, contactMatrix, Tf, Nf,
                Ns,
                int nc=30, double epsilon = 0.03,
                int tau_update_frequency = 1,
                verbose=False,
                method='deterministic',
                events=[],contactMatrices=[],
                events_repeat=False,
                events_subsequent=True,
                ):
        """
        Parameters
        ----------
        contactMatrix: python function(t), optional
             The social contact matrix C_{ij} denotes the 
             average number of contacts made per day by an 
             individual in class i with an individual in class j
             The default is None.
        Tf: float
            Final time of integrator.
        Nf: Int
            Number of time points to evaluate.
        Ns: int
            Number of samples of parameters to take.
        nc: int, optional
        epsilon: np.float64, optional
            Acceptable error in leap. The default is 0.03.
        tau_update_frequency : int, optional
            TODO
        verbose: bool, optional
            Verbosity of output. The default is False.
        Ti: float, optional
            Start time of integrator. The default is 0.
        method: str, optional
            Pyross integrator to use. The default is "deterministic".
        events: list of python functions, optional
            List of events that the current state can satisfy to change behaviour
            of the contact matrix. Event occurs when the value of the function
            changes sign. Event.direction determines which direction triggers
            the event, takign values {+1,-1}.
            The default is [].
        contactMatricies: list of python functions
            New contact matrix after the corresponding event occurs
            The default is [].
        events_repeat: bool, optional
            Wheither events is periodic in time. The default is false.
        events_subsequent : bool, optional
            TODO
             
        Returns
        -------
        out_dict : dict
            Dictionary containing the following keys:
                X: list
                    List of resultant trajectories
                t: list
                    List of times at which X is evaluated.
                X_mean : list
                    Mean trajectory of X
                X_std : list
                    Standard devation of trajectories of X at each time point.
                <init params> : Initial parameters passed at object instantiation.
                sample_parameters : list of parameters sampled to make trajectories.
                sample_inits : List of initial state vectors tried.
        """

        cdef:
            int M=self.M
            double [:] means_params=self.means_params, means_init=self.means_init
            double [:,:] cov_params=self.cov_params, cov_init=self.cov_init
            double start_time, end_time
            np.ndarray trajectories = np.zeros( [Ns,3*M,Nf] , dtype=DTYPE)
            np.ndarray mean_traj = np.zeros([ 3*M, Nf] ,dtype=DTYPE)
            np.ndarray std_traj = np.zeros([ 3*M, Nf] ,dtype=DTYPE)
            np.ndarray sample_parameters, sample_inits
            bint control = False

        if (contactMatrix == None) and (len(events) == 0):
            raise RuntimeError("Neither contactMatrix function nor list of events provided.")

        if len(events) > 0:
            control = True

        sample_parameters = np.random.multivariate_normal(means_params, cov_params, Ns)
        sample_inits = np.random.multivariate_normal(means_init, cov_init, Ns)

        start_time = timer()
        for i in range(Ns):
            if verbose:
                print('Running simulation {0} of {1}'.format(i+1,Ns),end='\r')
            while (sample_parameters[i] < 0).any():
                sample_parameters[i] = np.random.multivariate_normal(means_params,cov_params)
            while (sample_inits[i] < 0).any():
                sample_inits[i] = np.random.multivariate_normal(means_init, cov_init)
            parameters = {'fsa':self.fsa,
                        'alpha':sample_parameters[i,0],
                        'beta':sample_parameters[i,1],
                        'gIa':sample_parameters[i,2],'gIs':sample_parameters[i,3]}
            S0  = (sample_inits[i,    :M] * self.N).astype('int')
            Ia0 = (sample_inits[i, M  :2*M] * self.N).astype('int')
            Is0 = (sample_inits[i, 2*M:3*M] * self.N).astype('int')
            #
            if control:
                model = pyross.control.SIR(parameters, M, self.Ni)
                cur_result = model.simulate(S0, Ia0, Is0,
                              events, contactMatrices, Tf, Nf,
                              events_repeat=events_repeat,
                              events_subsequent=events_subsequent,
                              nc=nc,epsilon=epsilon,
                              tau_update_frequency=tau_update_frequency,
                              method=method)
            else:
                if method == 'deterministic':
                    model = pyross.deterministic.SIR(parameters, M, self.Ni)
                    cur_result = model.simulate(S0, Ia0, Is0,
                                    contactMatrix,
                                    Tf, Nf)
                else:
                    model = pyross.stochastic.SIR(parameters, M, self.Ni)
                    cur_result = model.simulate(S0, Ia0, Is0, contactMatrix, Tf, Nf,
                                  nc=nc,epsilon=epsilon,
                                  tau_update_frequency=tau_update_frequency,
                                  method=method)
            #
            trajectories[i] = cur_result['X'].T
        end_time = timer()
        if verbose:
            print('Finished. Time needed for evaluation:',time.strftime('%H:%M:%S',time.gmtime( end_time-start_time)) )
        #

        mean_traj = np.mean(trajectories,axis=0)
        std_traj = np.sqrt( np.var(trajectories,axis=0) )

        out_dict = {'X':trajectories, 't':cur_result['t'],
                    'X_mean':mean_traj,'X_std':std_traj,
                     'Ni':self.Ni, 'M':self.M,
                     'alpha':self.alpha, 'beta':self.beta,
                     'gIa':self.gIa, 'gIs':self.gIs,
                     'cov_params':self.cov_params,
                     'cov_init':self.cov_init,
                     'sample_parameters':sample_parameters,
                     'sample_inits':sample_inits
                      } #
        return out_dict


cdef class SEIR:
    """
    Susceptible, Exposed, Infected, Removed (SEIR)
    Ia: asymptomatic
    Is: symptomatic
    Parameters
    ----------
    parameters: dict
        Contains the following keys:
            alpha: float
                Estimate mean value of fraction of infected who are asymptomatic.
            beta: float
                Estimate mean value of rate of spread of infection.
            gIa: float
                Estimate mean value of rate of removal from asymptomatic individuals.
            gIs: float
                Estimate mean value of rate of removal from symptomatic individuals.
            fsa: float
                fraction by which symptomatic individuals do not self-isolate.
            gE: float
                Estimated mean value of rate of removal from exposed individuals.
            cov: np.array( )
                covariance matrix for all the estimated parameters.
    M: int
        Number of compartments of individual for each class.
        I.e len(contactMatrix)
    Ni: np.array(4*M, )
        Initial number in each compartment and class

    Methods
    -------
    simulate
    """
    cdef:
        readonly int N, M,
        readonly double alpha, beta, fsa, gIa, gIs, gE
        readonly int k_tot
        readonly np.ndarray rp0, Ni, drpdt, lld, CM, CC, means, cov

    def __init__(self, parameters, M, Ni):
        # these are the parameters we consider stochastic
        self.alpha = parameters.get('alpha')                    # fraction of asymptomatic infectives
        self.beta  = parameters.get('beta')                     # infection rate
        self.gE   = parameters.get('gE')                      # progression rate of E
        self.gIa   = parameters.get('gIa')                      # removal rate of Ia
        self.gIs   = parameters.get('gIs')                      # removal rate of Is
        #
        self.fsa   = parameters.get('fsa')                      # the self-isolation parameter
        #
        self.means = np.array([self.alpha,self.beta,self.gIa,self.gIs,self.gE],dtype=DTYPE)
        self.cov = parameters.get('cov') #np.array([[self.cov_beta_beta,self.cov_beta_gIa],
                  #      [self.cov_beta_gIa,self.cov_gIa_gIa]],
                  #      dtype=DTYPE)

        self.N     = np.sum(Ni)
        self.M     = M
        self.Ni    = np.zeros( self.M, dtype=DTYPE)             # # people in each age-group
        self.Ni    = Ni

        self.k_tot = 4


    def simulate(self, S0, E0, Ia0, Is0, contactMatrix, Tf, Nf,
                Ns,
                int nc=30, double epsilon = 0.03,
                int tau_update_frequency = 1,
                verbose=False,
                method='deterministic',
                events=[],contactMatrices=[],
                events_repeat=False,
                events_subsequent=True,
                ):
        """
        Parameters
        ----------
        S0: np.array(M,)
            Initial number of susceptables.
        E0: np.array(M,)
            Initial number of exposed.
        Ia0: np.array(M,)
            Initial number of asymptomatic infectives.
        Is0: np.array(M,)
            Initial number of symptomatic infectives.
        contactMatrix: python function(t), optional
             The social contact matrix C_{ij} denotes the 
             average number of contacts made per day by an 
             individual in class i with an individual in class j
             The default is None.
        Tf: float, optional
            Final time of integrator. The default is 100.
        Nf: Int, optional
            Number of time points to evaluate.The default is 101,
        Ns: int, optional
            Number of samples of parameters to take. The default is 1000.
        nc: int, optional
        epsilon: np.float64, optional
            Acceptable error in leap. The default is 0.03.
        tau_update_frequency : int, optional
        verbose: bool, optional
            Verbosity of output. The default is False.
        Ti: float, optional
            Start time of integrator. The default is 0.
        method: str, optional
            Pyross integrator to use. The default is "deterministic".
        events: list of python functions, optional
            List of events that the current state can satisfy to change behaviour
            of the contact matrix. Event occurs when the value of the function
            changes sign. Event.direction determines which direction triggers
            the event, takign values {+1,-1}.
            The default is [].
        contactMatricies: list of python functions
            New contact matrix after the corresponding event occurs
            The default is [].
        events_repeat: bool, optional
            Wheither events is periodic in time. The default is false.
        events_subsequent : bool, optional
            TODO
             
        Returns
        -------
        out_dict : dict
            Dictionary containing the following keys:
                X: list
                    List of resultant trajectories
                t: list
                    List of times at which X is evaluated.
                X_mean : list
                    Mean trajectory of X
                X_std : list
                    Standard devation of trajectories of X at each time point.
                <init params> : Initial parameters passed at object instantiation.
                sample_parameters : list of parameters sampled to make trajectories.
        """

        cdef:
            int M=self.M, k_tot = self.k_tot
            double [:] means  = self.means
            double [:,:] cov  = self.cov
            double start_time, end_time
            np.ndarray trajectories = np.zeros( [Ns,k_tot*M,Nf] , dtype=DTYPE)
            np.ndarray mean_traj = np.zeros([ k_tot*M, Nf] ,dtype=DTYPE)
            np.ndarray std_traj = np.zeros([ k_tot*M, Nf] ,dtype=DTYPE)
            np.ndarray sample_parameters
            bint control = False

        if (contactMatrix == None) and (len(events) == 0):
            raise RuntimeError("Neither contactMatrix function nor list of events provided.")

        if len(events) > 0:
            control = True

        sample_parameters = np.random.multivariate_normal(means, cov, Ns)

        start_time = timer()
        for i in range(Ns):
            if verbose:
                print('Running simulation {0} of {1}'.format(i+1,Ns),end='\r')
            while (sample_parameters[i] < 0).any():
                sample_parameters[i] = np.random.multivariate_normal(means,cov)
            parameters = { 'fsa':self.fsa,
                        'alpha':sample_parameters[i,0],
                        'beta':sample_parameters[i,1],
                        'gIa':sample_parameters[i,2],
                        'gIs':sample_parameters[i,3],
                        'gE':sample_parameters[i,4]}
            #
            if control:
                model = pyross.control.SEIR(parameters, M, self.Ni)
                cur_result = model.simulate(S0, E0, Ia0, Is0,
                              events, contactMatrices, Tf, Nf,
                              events_repeat=events_repeat,
                              events_subsequent=events_subsequent,
                              nc=nc,epsilon=epsilon,
                              tau_update_frequency=tau_update_frequency,
                              method=method)
            else:
                if method == 'deterministic':
                    model = pyross.deterministic.SEIR(parameters, M, self.Ni)
                    cur_result = model.simulate(S0, E0, Ia0, Is0,
                                    contactMatrix,
                                    Tf, Nf)
                else:
                    model = pyross.stochastic.SEIR(parameters, M, self.Ni)
                    cur_result = model.simulate(S0, E0, Ia0, Is0,
                                  contactMatrix, Tf, Nf,
                                  nc=nc,epsilon=epsilon,
                                  tau_update_frequency=tau_update_frequency,
                                  method=method)
            #
            trajectories[i] = cur_result['X'].T
        end_time = timer()
        if verbose:
            print('Finished. Time needed for evaluation:',time.strftime('%H:%M:%S',time.gmtime( end_time-start_time)) )
        #

        mean_traj = np.mean(trajectories,axis=0)
        std_traj = np.sqrt( np.var(trajectories,axis=0) )

        out_dict = {'X':trajectories, 't':cur_result['t'],
                    'X_mean':mean_traj,'X_std':std_traj,
                     'Ni':self.Ni, 'M':self.M,
                     'alpha':self.alpha, 'beta':self.beta,
                     'gE':self.gE,
                     'gIa':self.gIa, 'gIs':self.gIs,
                     'cov':self.cov,
                     'sample_parameters':sample_parameters
                      } #
        return out_dict


cdef class SEIR_latent:
    """
    Susceptible, Exposed, Infected, Removed (SEIR)
    Ia: asymptomatic
    Is: symptomatic
    
    Latent inference class to be used when observed data is incomplete.
    
    Parameters
    ----------
    parameters: dict
        Contains the following keys:
            alpha: float
                Estimate mean value of fraction of infected who are asymptomatic.
            beta: float
                Estimate mean value of rate of spread of infection.
            gIa: float
                Estimate mean value of rate of removal from asymptomatic individuals.
            gIs: float
                Estimate mean value of rate of removal from symptomatic individuals.
            fsa: float
                fraction by which symptomatic individuals do not self-isolate.
            gE: float
                Estimated mean value of rate of removal from exposed individuals.
            cov: np.array( )
                covariance matrix for all the estimated parameters.
            S0: np.array(M,)
                Estimate initial number of susceptables.
            E0: np.array(M,)
                Estimate initial number of exposed.
            Ia0: np.array(M,)
                Estimate initial number of asymptomatic infectives.
            Is0: np.array(M,)
                Estimate initial number of symptomatic infectives.
            cov_init : np.array((3*M, 3*M)) :
                Covariance matrix for the initial state.
    M: int
        Number of compartments of individual for each class.
        I.e len(contactMatrix)
    Ni: np.array(4*M, )
        Initial number in each compartment and class

    Methods
    -------
    simulate
    """
    cdef:
        readonly int N, M,
        readonly double alpha, beta, fsa, gIa, gIs, gE
        readonly int k_tot, k_random
        readonly np.ndarray rp0, Ni, drpdt, lld, CM, CC,
        readonly np.ndarray means_params, means_init, cov_params, cov_init

    def __init__(self, parameters, M, Ni):
        # these are the parameters we consider stochastic
        self.alpha = parameters.get('alpha')                    # fraction of asymptomatic infectives
        self.beta  = parameters.get('beta')                     # infection rate
        self.gE   = parameters.get('gE')                      # progression rate of E
        self.gIa   = parameters.get('gIa')                      # removal rate of Ia
        self.gIs   = parameters.get('gIs')                      # removal rate of Is
        #
        self.fsa   = parameters.get('fsa')                      # the self-isolation parameter
        #
        #
        self.cov_params = parameters.get('cov_params')
        self.cov_init = parameters.get('cov_init')
        #
        self.k_random = 5
        self.k_tot = 4
        self.means_params = np.array([self.alpha,self.beta,self.gIa,self.gIs,self.gE],dtype=DTYPE)
        self.means_init = np.zeros(self.k_tot*M,dtype=DTYPE)
        self.means_init[: M] = np.array(parameters.get('S0'),dtype=DTYPE)
        self.means_init[1*M:2*M] = np.array(parameters.get('E0'),dtype=DTYPE)
        self.means_init[2*M:3*M] = np.array(parameters.get('Ia0'),dtype=DTYPE)
        self.means_init[3*M:4*M] = np.array(parameters.get('Is0'),dtype=DTYPE)
        #
        self.N     = np.sum(Ni)
        self.M     = M
        self.Ni    = np.zeros( self.M, dtype=DTYPE)             # # people in each age-group
        self.Ni    = Ni
        #



    def simulate(self, contactMatrix, Tf, Nf,
                Ns,
                int nc=30, double epsilon = 0.03,
                int tau_update_frequency = 1,
                verbose=False,
                method='deterministic',
                events=[],contactMatrices=[],
                events_repeat=False,
                events_subsequent=True):
        """
        Parameters
        ----------
        contactMatrix: python function(t), optional
             The social contact matrix C_{ij} denotes the 
             average number of contacts made per day by an 
             individual in class i with an individual in class j
             The default is None.
        Tf: float, optional
            Final time of integrator. The default is 100.
        Nf: Int, optional
            Number of time points to evaluate.The default is 101,
        Ns: int, optional
            Number of samples of parameters to take. The default is 1000.
        nc: int, optional
        epsilon: np.float64, optional
            Acceptable error in leap. The default is 0.03.
        tau_update_frequency : int, optional
        verbose: bool, optional
            Verbosity of output. The default is False.
        Ti: float, optional
            Start time of integrator. The default is 0.
        method: str, optional
            Pyross integrator to use. The default is "deterministic".
        events: list of python functions, optional
            List of events that the current state can satisfy to change behaviour
            of the contact matrix. Event occurs when the value of the function
            changes sign. Event.direction determines which direction triggers
            the event, takign values {+1,-1}.
            The default is [].
        contactMatricies: list of python functions
            New contact matrix after the corresponding event occurs
            The default is [].
        events_repeat: bool, optional
            Wheither events is periodic in time. The default is false.
        events_subsequent : bool, optional
            TODO
             
        Returns
        -------
        out_dict : dict
            Dictionary containing the following keys:
                X: list
                    List of resultant trajectories
                t: list
                    List of times at which X is evaluated.
                X_mean : list
                    Mean trajectory of X
                X_std : list
                    Standard devation of trajectories of X at each time point.
                <init params> : Initial parameters passed at object instantiation.
                sample_parameters : list of parameters sampled to make trajectories.
                sample_inits : List of initial state vectors tried.
        """
        cdef:
            int M=self.M, k_tot = self.k_tot
            double [:] means_params=self.means_params, means_init=self.means_init
            double [:,:] cov_params=self.cov_params, cov_init=self.cov_init
            double start_time, end_time
            np.ndarray trajectories = np.zeros( [Ns,k_tot*M,Nf] , dtype=DTYPE)
            np.ndarray mean_traj = np.zeros([ k_tot*M, Nf] ,dtype=DTYPE)
            np.ndarray std_traj = np.zeros([ k_tot*M, Nf] ,dtype=DTYPE)
            np.ndarray sample_parameters, sample_inits
            bint control = False

        if (contactMatrix == None) and (len(events) == 0):
            raise RuntimeError("Neither contactMatrix function nor list of events provided.")

        if len(events) > 0:
            control = True

        sample_parameters = np.random.multivariate_normal(means_params, cov_params, Ns)
        sample_inits = np.random.multivariate_normal(means_init, cov_init, Ns)

        start_time = timer()
        for i in range(Ns):
            if verbose:
                print('Running simulation {0} of {1}'.format(i+1,Ns),end='\r')
            while (sample_parameters[i] < 0).any():
                sample_parameters[i] = np.random.multivariate_normal(means_params,cov_params)
            while (sample_inits[i] < 0).any():
                sample_inits[i] = np.random.multivariate_normal(means_init, cov_init)
            parameters = { 'fsa':self.fsa,
                        'alpha':sample_parameters[i,0],
                        'beta':sample_parameters[i,1],
                        'gIa':sample_parameters[i,2],
                        'gIs':sample_parameters[i,3],
                        'gE':sample_parameters[i,4]}
            S0  = ( sample_inits[i,   :1*M] * self.N ).astype('int')
            E0  = ( sample_inits[i,1*M:2*M] * self.N ).astype('int')
            Ia0 = ( sample_inits[i,2*M:3*M] * self.N ).astype('int')
            Is0 = ( sample_inits[i,3*M:4*M] * self.N ).astype('int')
            #
            if control:
                model = pyross.control.SEIR(parameters, M, self.Ni)
                cur_result = model.simulate(S0, E0, Ia0, Is0,
                              events, contactMatrices, Tf, Nf,
                              events_repeat=events_repeat,
                              events_subsequent=events_subsequent,
                              nc=nc,epsilon=epsilon,
                              tau_update_frequency=tau_update_frequency,
                              method=method)
            else:
                if method == 'deterministic':
                    model = pyross.deterministic.SEIR(parameters, M, self.Ni)
                    cur_result = model.simulate(S0, E0, Ia0, Is0,
                                    contactMatrix,
                                    Tf, Nf)
                else:
                    model = pyross.stochastic.SEIR(parameters, M, self.Ni)
                    cur_result = model.simulate(S0, E0, Ia0, Is0,
                                  contactMatrix, Tf, Nf,
                                  nc=nc,epsilon=epsilon,
                                  tau_update_frequency=tau_update_frequency,
                                  method=method)
            #
            trajectories[i] = cur_result['X'].T
        end_time = timer()
        if verbose:
            print('Finished. Time needed for evaluation:',time.strftime('%H:%M:%S',time.gmtime( end_time-start_time)) )
        #

        mean_traj = np.mean(trajectories,axis=0)
        std_traj = np.sqrt( np.var(trajectories,axis=0) )

        out_dict = {'X':trajectories, 't':cur_result['t'],
                    'X_mean':mean_traj,'X_std':std_traj,
                     'Ni':self.Ni, 'M':self.M,
                     'alpha':self.alpha, 'beta':self.beta,
                     'gE':self.gE,
                     'gIa':self.gIa, 'gIs':self.gIs,
                     'cov_params':self.cov_params,
                     'cov_init':self.cov_init,
                     'sample_parameters':sample_parameters,
                     'sample_inits':sample_inits
                      } #
        return out_dict


cdef class SEAIRQ():
    """
    Susceptible, Exposed, Infected, Removed (SEIR)
    Ia: asymptomatic
    Is: symptomatic
    A: Asymptomatic and infectious 
    Parameters
    ----------
    parameters: dict
        Contains the following keys:
            alpha: float
                Estimate mean value of fraction of infected who are asymptomatic.
            beta: float
                Estimate mean value of rate of spread of infection.
            gIa: float
                Estimate mean value of rate of removal from asymptomatic individuals.
            gIs: float
                Estimate mean value of rate of removal from symptomatic individuals.
            fsa: float
                fraction by which symptomatic individuals do not self-isolate.
            gE: float
                Estimated mean value of rate of removal from exposed individuals.
            gA: float
                Estimated mean value of rate of removal from activated individuals.
            cov: np.array( )
                covariance matrix for all the estimated parameters.
            tE  : float
                testing rate and contact tracing of exposeds
            tA  : float
                testing rate and contact tracing of activateds
            tIa: float
                testing rate and contact tracing of asymptomatics
            tIs: float
                testing rate and contact tracing of symptomatics
    M: int
        Number of compartments of individual for each class.
        I.e len(contactMatrix)
    Ni: np.array(4*M, )
        Initial number in each compartment and class

    Methods
    -------
    simulate
    """
    cdef:
        readonly int N, M,
        readonly double alpha, beta, gIa, gIs, gE, gA, fsa
        readonly double tE, tA, tIa, tIs
        readonly int k_tot
        readonly np.ndarray rp0, Ni, drpdt, lld, CM, CC, means, cov

    def __init__(self, parameters, M, Ni):
        # these are the parameters we consider stochastic
        self.alpha = parameters.get('alpha')                    # fraction of asymptomatic infectives
        self.beta  = parameters.get('beta')                     # infection rate
        self.gIa   = parameters.get('gIa')                      # removal rate of Ia
        self.gIs   = parameters.get('gIs')                      # removal rate of Is
        self.gE    = parameters.get('gE')                       # progression rate from E
        self.gA   = parameters.get('gA')                        # rate to go from A to Ia
        #
        #self.tS    = parameters.get('tS')                       # testing rate in S
        self.tE    = parameters.get('tE')                       # testing rate in E
        self.tA    = parameters.get('tA')                       # testing rate in A
        self.tIa   = parameters.get('tIa')                      # testing rate in Ia
        self.tIs   = parameters.get('tIs')                      # testing rate in Is

        # the following two parameters we consider to be exact
        self.fsa   = parameters.get('fsa')                      # the self-isolation parameter

        # vector of means & covariance matrix for Gaussian distribution
        self.means = np.array([self.alpha,self.beta,
                                self.gIa,self.gIs,
                                self.gE,self.gA],dtype=DTYPE)
        self.cov = parameters.get('cov') # 11x11 matrix

        self.N     = np.sum(Ni)
        self.M     = M
        self.Ni    = np.zeros( self.M, dtype=DTYPE)             # # people in each age-group
        self.Ni    = Ni

        self.k_tot = 6 # total number of explicit states per age group


    def simulate(self, S0, E0, A0, Ia0, Is0, Q0,
                contactMatrix, Tf, Nf,
                Ns,
                int nc=30, double epsilon = 0.03,
                int tau_update_frequency = 1,
                verbose=False,
                method='deterministic',
                events=[],contactMatrices=[],
                events_repeat=False,
                events_subsequent=True):
        """
        Parameters
        ----------
        S0: np.array
            Initial number of susceptables.
        E0: np.array
            Initial number of exposeds.
        A0: np.array
            Initial number of activateds.
        Ia0: np.array
            Initial number of asymptomatic infectives.
        Is0: np.array
            Initial number of symptomatic infectives.
        Q0: np.array
            Initial number of quarantineds.
        contactMatrix: python function(t), optional
             The social contact matrix C_{ij} denotes the 
             average number of contacts made per day by an 
             individual in class i with an individual in class j
             The default is None.
        Tf: float, optional
            Final time of integrator. The default is 100.
        Nf: Int, optional
            Number of time points to evaluate.The default is 101,
        Ns: int, optional
            Number of samples of parameters to take. The default is 1000.
        nc: int, optional
        epsilon: np.float64, optional
            Acceptable error in leap. The default is 0.03.
        tau_update_frequency : int, optional
        verbose: bool, optional
            Verbosity of output. The default is False.
        Ti: float, optional
            Start time of integrator. The default is 0.
        method: str, optional
            Pyross integrator to use. The default is "deterministic".
        events: list of python functions, optional
            List of events that the current state can satisfy to change behaviour
            of the contact matrix. Event occurs when the value of the function
            changes sign. Event.direction determines which direction triggers
            the event, takign values {+1,-1}.
            The default is [].
        contactMatricies: list of python functions
            New contact matrix after the corresponding event occurs
            The default is [].
        events_repeat: bool, optional
            Wheither events is periodic in time. The default is false.
        events_subsequent : bool, optional
            TODO
             
        Returns
        -------
        out_dict : dict
            Dictionary containing the following keys:
                X: list
                    List of resultant trajectories
                t: list
                    List of times at which X is evaluated.
                X_mean : list
                    Mean trajectory of X
                X_std : list
                    Standard devation of trajectories of X at each time point.
                <init params> : Initial parameters passed at object instantiation.
                sample_parameters : list of parameters sampled to make trajectories.
        """

        cdef:
            int M=self.M, k_tot = self.k_tot
            double [:] means  = self.means
            double [:,:] cov  = self.cov
            double start_time, end_time
            np.ndarray trajectories = np.zeros( [Ns,k_tot*M,Nf] , dtype=DTYPE)
            np.ndarray mean_traj = np.zeros([ k_tot*M, Nf] ,dtype=DTYPE)
            np.ndarray std_traj = np.zeros([ k_tot*M, Nf] ,dtype=DTYPE)
            np.ndarray sample_parameters
            bint control = False

        if (contactMatrix == None) and (len(events) == 0):
            raise RuntimeError("Neither contactMatrix function nor list of events provided.")

        if len(events) > 0:
            control = True

        sample_parameters = np.random.multivariate_normal(means, cov, Ns)

        start_time = timer()
        for i in range(Ns):
            if verbose:
                print('Running simulation {0} of {1}'.format(i+1,Ns),end='\r')
            while (sample_parameters[i] < 0).any():
                sample_parameters[i] = np.random.multivariate_normal(means,cov)
            parameters = { 'fsa':self.fsa,
                        'alpha':sample_parameters[i,0],
                        'beta':sample_parameters[i,1],
                        'gIa':sample_parameters[i,2],
                        'gIs':sample_parameters[i,3],
                        'gE':sample_parameters[i,4],
                        'gA':sample_parameters[i,5],
                        'tE':self.tE,
                        'tA':self.tA,
                        'tIa':self.tIa,
                        'tIs':self.tIs}
            #
            if control:
                model = pyross.control.SEAIRQ(parameters, M, self.Ni)
                cur_result = model.simulate(S0, E0, A0, Ia0, Is0, Q0,
                              events, contactMatrices, Tf, Nf,
                              events_repeat=events_repeat,
                              events_subsequent=events_subsequent,
                              nc=nc,epsilon=epsilon,
                              tau_update_frequency=tau_update_frequency,
                              method=method)
            else:
                if method == 'deterministic':
                    model = pyross.deterministic.SEAIRQ(parameters, M, self.Ni)
                    cur_result = model.simulate(S0, E0, A0, Ia0, Is0, Q0,
                                    contactMatrix,
                                    Tf, Nf)
                else:
                    model = pyross.stochastic.SEAIRQ(parameters, M, self.Ni)
                    cur_result = model.simulate(S0, E0, A0, Ia0, Is0, Q0,
                                  contactMatrix, Tf, Nf,
                                  nc=nc,epsilon=epsilon,
                                  tau_update_frequency=tau_update_frequency,
                                  method=method)
            #
            trajectories[i] = cur_result['X'].T
        end_time = timer()
        if verbose:
            print('Finished. Time needed for evaluation:',time.strftime('%H:%M:%S',time.gmtime( end_time-start_time)) )
        #

        mean_traj = np.mean(trajectories,axis=0)
        std_traj = np.sqrt( np.var(trajectories,axis=0) )

        out_dict={'X':trajectories, 't':cur_result['t'],
                    'X_mean':mean_traj,'X_std':std_traj,
                'Ni':self.Ni, 'M':self.M,
                'alpha':self.alpha,'beta':self.beta,
                'gIa':self.gIa,'gIs':self.gIs,
                'gE':self.gE,'gA':self.gA,
                'tE':self.tE,'tIa':self.tIa,'tIs':self.tIs,
                'cov':self.cov,
                'sample_parameters':sample_parameters}
        return out_dict


cdef class SEAIRQ_latent():
    """
    Susceptible, Exposed, Infected, Removed (SEIR)
    Ia: asymptomatic
    Is: symptomatic
    A: Asymptomatic and infectious 
    
    Latent inference class to be used when observed data is incomplete.
    Parameters
    ----------
    parameters: dict
        Contains the following keys:
            alpha: float
                Estimate mean value of fraction of infected who are asymptomatic.
            beta: float
                Estimate mean value of rate of spread of infection.
            gIa: float
                Estimate mean value of rate of removal from asymptomatic individuals.
            gIs: float
                Estimate mean value of rate of removal from symptomatic individuals.
            fsa: float
                fraction by which symptomatic individuals do not self-isolate.
            gE: float
                Estimated mean value of rate of removal from exposed individuals.
            gA: float
                Estimated mean value of rate of removal from activated individuals.
            cov: np.array( )
                covariance matrix for all the estimated parameters.
            tE  : float
                testing rate and contact tracing of exposeds
            tA  : float
                testing rate and contact tracing of activateds
            tIa: float
                testing rate and contact tracing of asymptomatics
            tIs: float
                testing rate and contact tracing of symptomatics
            S0: np.array(M,)
                Estimate initial number of susceptables.
            E0: np.array(M,)
                Estimate initial number of exposed.
            A0: np.array(M,)
                Estimate initial number of activated.
            Ia0: np.array(M,)
                Estimate initial number of asymptomatic infectives.
            Is0: np.array(M,)
                Estimate initial number of symptomatic infectives.
            Q0: np.array(M,)
                Estimate initial number of quarantined.
            cov_init : np.array((3*M, 3*M)) :
                Covariance matrix for the initial state.
    M: int
        Number of compartments of individual for each class.
        I.e len(contactMatrix)
    Ni: np.array(4*M, )
        Initial number in each compartment and class

    Methods
    -------
    simulate
    """
    cdef:
        readonly int N, M
        readonly double alpha, beta, gIa, gIs, gE, gA, fsa
        readonly double tE, tA, tIa, tIs
        readonly int k_tot, k_random
        readonly np.ndarray rp0, Ni, drpdt, lld, CM, CC, means, cov
        readonly np.ndarray means_params, means_init, cov_params, cov_init

    def __init__(self, parameters, M, Ni):
        # these are the parameters we consider stochastic
        self.alpha = parameters.get('alpha')                    # fraction of asymptomatic infectives
        self.beta  = parameters.get('beta')                     # infection rate
        self.gIa   = parameters.get('gIa')                      # removal rate of Ia
        self.gIs   = parameters.get('gIs')                      # removal rate of Is
        self.gE    = parameters.get('gE')                       # progression rate from E
        self.gA   = parameters.get('gA')                        # rate to go from A to Ia
        #
        #self.tS    = parameters.get('tS')                       # testing rate in S
        self.tE    = parameters.get('tE')                       # testing rate in E
        self.tA    = parameters.get('tA')                       # testing rate in A
        self.tIa   = parameters.get('tIa')                      # testing rate in Ia
        self.tIs   = parameters.get('tIs')                      # testing rate in Is

        # the following two parameters we consider to be exact
        self.fsa   = parameters.get('fsa')                      # the self-isolation parameter

        self.k_tot = 6 # total number of explicit states per age group
        #
        self.cov_params = parameters.get('cov_params')
        self.cov_init = parameters.get('cov_init')
        #
        self.k_random = 6
        self.means_params = np.array([self.alpha,self.beta,
                          self.gIa,self.gIs,self.gE,self.gA],dtype=DTYPE)
        self.means_init = np.zeros(self.k_tot*M,dtype=DTYPE)
        self.means_init[: M] = np.array(parameters.get('S0'),dtype=DTYPE)
        self.means_init[1*M:2*M] = np.array(parameters.get('E0'),dtype=DTYPE)
        self.means_init[2*M:3*M] = np.array(parameters.get('A0'),dtype=DTYPE)
        self.means_init[3*M:4*M] = np.array(parameters.get('Ia0'),dtype=DTYPE)
        self.means_init[4*M:5*M] = np.array(parameters.get('Is0'),dtype=DTYPE)
        self.means_init[5*M:6*M] = np.array(parameters.get('Q0'),dtype=DTYPE)

        self.N     = np.sum(Ni)
        self.M     = M
        self.Ni    = np.zeros( self.M, dtype=DTYPE)             # # people in each age-group
        self.Ni    = Ni



    def simulate(self,
                contactMatrix, Tf, Nf,
                Ns,
                int nc=30, double epsilon = 0.03,
                int tau_update_frequency = 1,
                verbose=False,
                method='deterministic',
                events=[],contactMatrices=[],
                events_repeat=False,
                events_subsequent=True):
        """
        Parameters
        ----------
        contactMatrix: python function(t), optional
             The social contact matrix C_{ij} denotes the 
             average number of contacts made per day by an 
             individual in class i with an individual in class j
             The default is None.
        Tf: float, optional
            Final time of integrator. The default is 100.
        Nf: Int, optional
            Number of time points to evaluate.The default is 101,
        Ns: int, optional
            Number of samples of parameters to take. The default is 1000.
        nc: int, optional
        epsilon: np.float64, optional
            Acceptable error in leap. The default is 0.03.
        tau_update_frequency : int, optional
        verbose: bool, optional
            Verbosity of output. The default is False.
        Ti: float, optional
            Start time of integrator. The default is 0.
        method: str, optional
            Pyross integrator to use. The default is "deterministic".
        events: list of python functions, optional
            List of events that the current state can satisfy to change behaviour
            of the contact matrix. Event occurs when the value of the function
            changes sign. Event.direction determines which direction triggers
            the event, takign values {+1,-1}.
            The default is [].
        contactMatricies: list of python functions
            New contact matrix after the corresponding event occurs
            The default is [].
        events_repeat: bool, optional
            Wheither events is periodic in time. The default is false.
        events_subsequent : bool, optional
            TODO
             
        Returns
        -------
        out_dict : dict
            Dictionary containing the following keys:
                X: list
                    List of resultant trajectories
                t: list
                    List of times at which X is evaluated.
                X_mean : list
                    Mean trajectory of X
                X_std : list
                    Standard devation of trajectories of X at each time point.
                <init params> : Initial parameters passed at object instantiation.
                sample_parameters : list of parameters sampled to make trajectories.
                sample_inits : List of initial state vectors tried.
        """

        cdef:
            int M=self.M, k_tot = self.k_tot
            double [:] means_params=self.means_params, means_init=self.means_init
            double [:,:] cov_params=self.cov_params, cov_init=self.cov_init
            double start_time, end_time
            np.ndarray trajectories = np.zeros( [Ns,k_tot*M,Nf] , dtype=DTYPE)
            np.ndarray mean_traj = np.zeros([ k_tot*M, Nf] ,dtype=DTYPE)
            np.ndarray std_traj = np.zeros([ k_tot*M, Nf] ,dtype=DTYPE)
            np.ndarray sample_parameters, sample_inits
            bint control = False

        if (contactMatrix == None) and (len(events) == 0):
            raise RuntimeError("Neither contactMatrix function nor list of events provided.")

        if len(events) > 0:
            control = True

        sample_parameters = np.random.multivariate_normal(means_params, cov_params, Ns)
        sample_inits = np.random.multivariate_normal(means_init, cov_init, Ns)

        start_time = timer()
        for i in range(Ns):
            if verbose:
                print('Running simulation {0} of {1}'.format(i+1,Ns),end='\r')
            while (sample_parameters[i] < 0).any():
                sample_parameters[i] = np.random.multivariate_normal(means_params,cov_params)
            while (sample_inits[i] < 0).any():
                sample_inits[i] = np.random.multivariate_normal(means_init, cov_init)
            parameters = { 'fsa':self.fsa,
                        'alpha':sample_parameters[i,0],
                        'beta':sample_parameters[i,1],
                        'gIa':sample_parameters[i,2],
                        'gIs':sample_parameters[i,3],
                        'gE':sample_parameters[i,4],
                        'gA':sample_parameters[i,5],
                        'tE':self.tE,
                        'tA':self.tA,
                        'tIa':self.tIa,
                        'tIs':self.tIs}
            S0 =  (sample_inits[i,: M] * self.N).astype('int')
            E0 =  (sample_inits[i,M: 2*M]* self.N).astype('int')
            A0 =  (sample_inits[i,2*M: 3*M]* self.N).astype('int')
            Ia0 =  (sample_inits[i,3*M: 4*M]* self.N).astype('int')
            Is0 =  (sample_inits[i,4*M: 5*M]* self.N).astype('int')
            Q0 =  (sample_inits[i,5*M: 6*M]* self.N).astype('int')
            #
            if control:
                model = pyross.control.SEAIRQ(parameters, M, self.Ni)
                cur_result = model.simulate(S0, E0, A0, Ia0, Is0, Q0,
                              events, contactMatrices, Tf, Nf,
                              events_repeat=events_repeat,
                              events_subsequent=events_subsequent,
                              nc=nc,epsilon=epsilon,
                              tau_update_frequency=tau_update_frequency,
                              method=method)
            else:
                if method == 'deterministic':
                    model = pyross.deterministic.SEAIRQ(parameters, M, self.Ni)
                    cur_result = model.simulate(S0, E0, A0, Ia0, Is0, Q0,
                                    contactMatrix,
                                    Tf, Nf)
                else:
                    model = pyross.stochastic.SEAIRQ(parameters, M, self.Ni)
                    cur_result = model.simulate(S0, E0, A0, Ia0, Is0, Q0,
                                  contactMatrix, Tf, Nf,
                                  nc=nc,epsilon=epsilon,
                                  tau_update_frequency=tau_update_frequency,
                                  method=method)
            #
            trajectories[i] = cur_result['X'].T
        end_time = timer()
        if verbose:
            print('Finished. Time needed for evaluation:',time.strftime('%H:%M:%S',time.gmtime( end_time-start_time)) )
        #

        mean_traj = np.mean(trajectories,axis=0)
        std_traj = np.sqrt( np.var(trajectories,axis=0) )

        out_dict={'X':trajectories, 't':cur_result['t'],
                    'X_mean':mean_traj,'X_std':std_traj,
                'Ni':self.Ni, 'M':self.M,
                'alpha':self.alpha,'beta':self.beta,
                'gIa':self.gIa,'gIs':self.gIs,
                'gE':self.gE,'gA':self.gA,
                'tE':self.tE,'tIa':self.tIa,'tIs':self.tIs,
                'cov_params':self.cov_params,
                'cov_init':self.cov_init,
                'sample_parameters':sample_parameters,
                'sample_inits':sample_inits}
        return out_dict


cdef class SEAI5R():
    """
    Susceptible, Exposed, Activates, Infected, Removed (SEAIR)
    The infected class has 5 groups:
    * Ia: asymptomatic
    * Is: symptomatic
    * Ih: hospitalized
    * Ic: ICU
    * Im: Mortality

    S  ---> E
    E  ---> Ia, Is
    Ia ---> R
    Is ---> Ih, R
    Ih ---> Ic, R
    Ic ---> Im, R
    Parameters
    ----------
    parameters: dict
        Contains the following keys:
            alpha: float
                Estimate mean value of fraction of infected who are asymptomatic.
            beta: float
                Estimate mean value of rate of spread of infection.
            gIa: float
                Estimate mean value of rate of removal from asymptomatic individuals.
            gIs: float
                Estimate mean value of rate of removal from symptomatic individuals.
            fsa: float
                fraction by which symptomatic individuals do not self-isolate.
            gE: float
                Estimate mean value of rate of removal from exposeds individuals.
            gA: float
                Estimate mean value of rate of removal from activated individuals.
            gIh: float
                Rate of hospitalisation of infected individuals.
            gIc: float
                Rate hospitalised individuals are moved to intensive care.
            cov: np.array
                covariance matrix for all the estimated parameters.
            sa: float, np.array (M,)
                daily arrival of new susceptables.
                sa is rate of additional/removal of population by birth etc
            hh: float, np.array (M,)
                fraction hospitalised from Is
            cc: float, np.array (M,)
                fraction sent to intensive care from hospitalised.
            mm: float, np.array (M,)
                mortality rate in intensive care
    M: int
        Number of compartments of individual for each class.
        I.e len(contactMatrix)
    Ni: np.array(9*M, )
        Initial number in each compartment and class

    Methods
    -------
    simulate
    """
    cdef:
        readonly int N, M,
        readonly double alpha, beta, gE, gA, gIa, gIs, gIh, gIc, fsa, fh
        readonly int k_tot
        readonly np.ndarray rp0, Ni, drpdt, lld, CM, CC, sa, hh, cc, mm, means, cov

    def __init__(self, parameters, M, Ni):
        # these are the parameters we consider stochastic
        self.alpha = parameters.get('alpha')                    # fraction of asymptomatic infectives
        self.beta  = parameters.get('beta')                     # infection rate
        self.gE    = parameters.get('gE')                       # progression rate of E class
        self.gA    = parameters.get('gA')                       # progression rate of A class
        self.gIa   = parameters.get('gIa')                      # removal rate of Ia
        self.gIs   = parameters.get('gIs')                      # removal rate of Is

        # the following parameters we consider to be exact
        self.fsa   = parameters.get('fsa')                      # the self-isolation parameter
        self.fh    = parameters.get('fh')                       # the self-isolation parameter of hospitalizeds
        self.gIh   = parameters.get('gIh')                      # removal rate of Ih
        self.gIc   = parameters.get('gIc')                      # removal rate of Ic
        #
        sa         = parameters.get('sa')                       # daily arrival of new susceptibles
        hh         = parameters.get('hh')                       # hospital
        cc         = parameters.get('cc')                       # ICU
        mm         = parameters.get('mm')                       # mortality
        #iaa        = parameters.get('iaa')                      # daily arrival of new asymptomatics

        self.sa    = np.zeros( self.M, dtype = DTYPE)
        if np.size(sa)==1:
            self.sa = sa*np.ones(M)
        elif np.size(sa)==M:
            self.sa= sa
        else:
            print('sa can be a number or an array of size M')

        self.hh    = np.zeros( self.M, dtype = DTYPE)
        if np.size(hh)==1:
            self.hh = hh*np.ones(M)
        elif np.size(hh)==M:
            self.hh= hh
        else:
            print('hh can be a number or an array of size M')

        self.cc    = np.zeros( self.M, dtype = DTYPE)
        if np.size(cc)==1:
            self.cc = cc*np.ones(M)
        elif np.size(cc)==M:
            self.cc= cc
        else:
            print('cc can be a number or an array of size M')

        self.mm    = np.zeros( self.M, dtype = DTYPE)
        if np.size(mm)==1:
            self.mm = mm*np.ones(M)
        elif np.size(mm)==M:
            self.mm= mm
        else:
            print('mm can be a number or an array of size M')

        # vector of means & covariance matrix for Gaussian distribution
        self.means = np.array([self.alpha,
                                self.beta,
                                self.gIa,self.gIs,
                                self.gE,self.gA],
                                #self.gIh,self.gIc],
                                dtype=DTYPE)
        self.cov = parameters.get('cov') # 7x7 matrix

        self.N     = np.sum(Ni)
        self.M     = M
        self.Ni    = np.zeros( self.M, dtype=DTYPE)             # # people in each age-group
        self.Ni    = Ni

        self.k_tot = 9 # total number of explicit states per age group


    def simulate(self, S0, E0, A0, Ia0, Is0, Ih0, Ic0, Im0,
                contactMatrix, Tf, Nf,
                Ns,
                int nc=30, double epsilon = 0.03,
                int tau_update_frequency = 1,
                verbose=False,
                method='deterministic',
                events=[],contactMatrices=[],
                events_repeat=False,
                events_subsequent=True):
        """
        Parameters
        ----------
        S0: np.array(M,)
            Initial number of susceptables.
        E0: np.array(M,)
            Initial number of exposed.
        A0: np.array(M,)
            Initial number of activated.
        Ia0: np.array(M,)
            Initial number of asymptomatic infectives.
        Is0: np.array(M,)
            Initial number of symptomatic infectives.
        Ih0: np.array(M,)
            Initial number of hospitalised.
        Ic0: np.array(M,)
            Initial number in intensive care.
        Im0: np.array(M,)
            Initial number of mortalities.
        contactMatrix: python function(t), optional
             The social contact matrix C_{ij} denotes the 
             average number of contacts made per day by an 
             individual in class i with an individual in class j
             The default is None.
        Tf: float, optional
            Final time of integrator. The default is 100.
        Nf: Int, optional
            Number of time points to evaluate.The default is 101,
        Ns: int, optional
            Number of samples of parameters to take. The default is 1000.
        nc: int, optional
        epsilon: np.float64, optional
            Acceptable error in leap. The default is 0.03.
        tau_update_frequency : int, optional
        verbose: bool, optional
            Verbosity of output. The default is False.
        Ti: float, optional
            Start time of integrator. The default is 0.
        method: str, optional
            Pyross integrator to use. The default is "deterministic".
        events: list of python functions, optional
            List of events that the current state can satisfy to change behaviour
            of the contact matrix. Event occurs when the value of the function
            changes sign. Event.direction determines which direction triggers
            the event, takign values {+1,-1}.
            The default is [].
        contactMatricies: list of python functions
            New contact matrix after the corresponding event occurs
            The default is [].
        events_repeat: bool, optional
            Wheither events is periodic in time. The default is false.
        events_subsequent : bool, optional
            TODO
             
        Returns
        -------
        out_dict : dict
            Dictionary containing the following keys:
                X: list
                    List of resultant trajectories
                t: list
                    List of times at which X is evaluated.
                X_mean : list
                    Mean trajectory of X
                X_std : list
                    Standard devation of trajectories of X at each time point.
                <init params> : Initial parameters passed at object instantiation.
                sample_parameters : list of parameters sampled to make trajectories.
        """

        cdef:
            int M=self.M, k_tot = self.k_tot
            double [:] means  = self.means
            double [:,:] cov  = self.cov
            double start_time, end_time
            np.ndarray trajectories = np.zeros( [Ns,k_tot*M,Nf] , dtype=DTYPE)
            np.ndarray mean_traj = np.zeros([ k_tot*M, Nf] ,dtype=DTYPE)
            np.ndarray std_traj = np.zeros([ k_tot*M, Nf] ,dtype=DTYPE)
            np.ndarray sample_parameters
            bint control = False

        if (contactMatrix == None) and (len(events) == 0):
            raise RuntimeError("Neither contactMatrix function nor list of events provided.")

        if len(events) > 0:
            control = True

        sample_parameters = np.random.multivariate_normal(means, cov, Ns)

        start_time = timer()
        for i in range(Ns):
            if verbose:
                print('Running simulation {0} of {1}'.format(i+1,Ns),end='\r')
            while (sample_parameters[i] < 0).any():
                sample_parameters[i] = np.random.multivariate_normal(means,cov)
            parameters = {'alpha':sample_parameters[i,0],
                        'beta':sample_parameters[i,1],
                        'gIa':sample_parameters[i,2],
                        'gIs':sample_parameters[i,3],
                        'gE':sample_parameters[i,4],
                        'gA':sample_parameters[i,5],
                        'gIh':self.gIh,'gIc':self.gIc,
                        'fsa':self.fsa,'fh':self.fh,
                        'sa':self.sa,'hh':self.hh,
                        'mm':self.mm,'cc':self.cc
                          }
            #
            if control:
                model = pyross.control.SEAI5R(parameters, M, self.Ni)
                cur_result = model.simulate(S0, E0, A0, Ia0, Is0, Ih0, Ic0, Im0,
                              events, contactMatrices, Tf, Nf,
                              events_repeat=events_repeat,
                              events_subsequent=events_subsequent,
                              nc=nc,epsilon=epsilon,
                              tau_update_frequency=tau_update_frequency,
                              method=method)
            else:
                if method == 'deterministic':
                    model = pyross.deterministic.SEAI5R(parameters, M, self.Ni)
                    cur_result = model.simulate(S0, E0, A0, Ia0, Is0, Ih0, Ic0, Im0,
                                    contactMatrix,
                                    Tf, Nf)
                else:
                    model = pyross.stochastic.SEAI5R(parameters, M, self.Ni)
                    cur_result = model.simulate(S0, E0, A0, Ia0, Is0, Ih0, Ic0, Im0,
                                  contactMatrix, Tf, Nf,
                                  nc=nc,epsilon=epsilon,
                                  tau_update_frequency=tau_update_frequency,
                                  method=method)
            #
            trajectories[i] = cur_result['X'].T
        end_time = timer()
        if verbose:
            print('Finished. Time needed for evaluation:',time.strftime('%H:%M:%S',time.gmtime( end_time-start_time)) )
        #

        mean_traj = np.mean(trajectories,axis=0)
        std_traj = np.sqrt( np.var(trajectories,axis=0) )

        out_dict = {'X':trajectories, 't':cur_result['t'],
                    'X_mean':mean_traj,'X_std':std_traj,
                      'Ni':self.Ni, 'M':self.M,
                      'alpha':self.alpha, 'beta':self.beta,
                      'gIa':self.gIa,'gIs':self.gIs,
                      'gIh':self.gIh,'gIc':self.gIc,
                      'fsa':self.fsa,'fh':self.fh,
                      'gE':self.gE,'gA':self.gA,
                      'sa':self.sa,'hh':self.hh,
                      'mm':self.mm,'cc':self.cc,
                      'cov':self.cov,
                      'sample_parameters':sample_parameters
                      }
        return out_dict


cdef class SEAI5R_latent():
    """
    Susceptible, Exposed, Activates, Infected, Removed (SEAIR)
    The infected class has 5 groups:
    * Ia: asymptomatic
    * Is: symptomatic
    * Ih: hospitalized
    * Ic: ICU
    * Im: Mortality

    S  ---> E
    E  ---> Ia, Is
    Ia ---> R
    Is ---> Ih, R
    Ih ---> Ic, R
    Ic ---> Im, R
    
    Latent inference class to be used when observed data is incomplete.
    
    Parameters
    ----------
    parameters: dict
        Contains the following keys:
            alpha: float
                Estimate mean value of fraction of infected who are asymptomatic.
            beta: float
                Estimate mean value of rate of spread of infection.
            gIa: float
                Estimate mean value of rate of removal from asymptomatic individuals.
            gIs: float
                Estimate mean value of rate of removal from symptomatic individuals.
            fsa: float
                fraction by which symptomatic individuals do not self-isolate.
            gE: float
                Estimate mean value of rate of removal from exposeds individuals.
            gA: float
                Estimate mean value of rate of removal from activated individuals.
            gIh: float
                Rate of hospitalisation of infected individuals.
            gIc: float
                Rate hospitalised individuals are moved to intensive care.
            cov: np.array
                covariance matrix for all the estimated parameters.
            sa: float, np.array (M,)
                daily arrival of new susceptables.
                sa is rate of additional/removal of population by birth etc
            hh: float, np.array (M,)
                fraction hospitalised from Is
            cc: float, np.array (M,)
                fraction sent to intensive care from hospitalised.
            mm: float, np.array (M,)
                mortality rate in intensive care
    M: int
        Number of compartments of individual for each class.
        I.e len(contactMatrix)
    Ni: np.array(9*M, )
        Initial number in each compartment and class

    Methods
    -------
    simulate
    """
    cdef:
        readonly int N, M,
        readonly double alpha, beta, gE, gA, gIa, gIs, gIh, gIc, fsa, fh
        readonly int k_random, k_tot
        readonly np.ndarray rp0, Ni, drpdt, lld, CM, CC, sa, hh, cc, mm
        readonly np.ndarray means_params, means_init, cov_params, cov_init


    def __init__(self, parameters, M, Ni):
        # these are the parameters we consider stochastic
        self.alpha = parameters.get('alpha')                    # fraction of asymptomatic infectives
        self.beta  = parameters.get('beta')                     # infection rate
        self.gE    = parameters.get('gE')                       # progression rate of E class
        self.gA    = parameters.get('gA')                       # progression rate of A class
        self.gIa   = parameters.get('gIa')                      # removal rate of Ia
        self.gIs   = parameters.get('gIs')                      # removal rate of Is

        # the following parameters we consider to be exact
        self.fsa   = parameters.get('fsa')                      # the self-isolation parameter
        self.fh    = parameters.get('fh')                       # the self-isolation parameter of hospitalizeds
        self.gIh   = parameters.get('gIh')                      # removal rate of Ih
        self.gIc   = parameters.get('gIc')                      # removal rate of Ic
        #
        sa         = parameters.get('sa')                       # daily arrival of new susceptibles
        hh         = parameters.get('hh')                       # hospital
        cc         = parameters.get('cc')                       # ICU
        mm         = parameters.get('mm')                       # mortality
        #iaa        = parameters.get('iaa')                      # daily arrival of new asymptomatics

        self.sa    = np.zeros( self.M, dtype = DTYPE)
        if np.size(sa)==1:
            self.sa = sa*np.ones(M)
        elif np.size(sa)==M:
            self.sa= sa
        else:
            print('sa can be a number or an array of size M')

        self.hh    = np.zeros( self.M, dtype = DTYPE)
        if np.size(hh)==1:
            self.hh = hh*np.ones(M)
        elif np.size(hh)==M:
            self.hh= hh
        else:
            print('hh can be a number or an array of size M')

        self.cc    = np.zeros( self.M, dtype = DTYPE)
        if np.size(cc)==1:
            self.cc = cc*np.ones(M)
        elif np.size(cc)==M:
            self.cc= cc
        else:
            print('cc can be a number or an array of size M')

        self.mm    = np.zeros( self.M, dtype = DTYPE)
        if np.size(mm)==1:
            self.mm = mm*np.ones(M)
        elif np.size(mm)==M:
            self.mm= mm
        else:
            print('mm can be a number or an array of size M')

        self.k_tot = 8 # total number of explicit states per age group
        #
        self.cov_params = parameters.get('cov_params')
        self.cov_init = parameters.get('cov_init')
        #
        self.k_random = 6
        self.means_params = np.array([self.alpha,self.beta,
                          self.gIa,self.gIs,self.gE,self.gA],dtype=DTYPE)
        self.means_init = np.zeros(self.k_tot*M,dtype=DTYPE)
        self.means_init[: M] = np.array(parameters.get('S0'),dtype=DTYPE)
        self.means_init[1*M:2*M] = np.array(parameters.get('E0'),dtype=DTYPE)
        self.means_init[2*M:3*M] = np.array(parameters.get('A0'),dtype=DTYPE)
        self.means_init[3*M:4*M] = np.array(parameters.get('Ia0'),dtype=DTYPE)
        self.means_init[4*M:5*M] = np.array(parameters.get('Is0'),dtype=DTYPE)
        self.means_init[5*M : 6*M] = np.array(parameters.get('Ih0'),dtype=DTYPE)
        self.means_init[6*M : 7*M] = np.array(parameters.get('Ic0'),dtype=DTYPE)
        self.means_init[7*M : 8*M] = np.array(parameters.get('Im0'),dtype=DTYPE)

        self.N     = np.sum(Ni)
        self.M     = M
        self.Ni    = np.zeros( self.M, dtype=DTYPE)             # # people in each age-group
        self.Ni    = Ni



    def simulate(self,
                contactMatrix, Tf, Nf,
                Ns,
                int nc=30, double epsilon = 0.03,
                int tau_update_frequency = 1,
                verbose=False,
                method='deterministic',
                events=[],contactMatrices=[],
                events_repeat=False,
                events_subsequent=True):
        """
        Parameters
        ----------
        contactMatrix: python function(t), optional
             The social contact matrix C_{ij} denotes the 
             average number of contacts made per day by an 
             individual in class i with an individual in class j
             The default is None.
        Tf: float, optional
            Final time of integrator. The default is 100.
        Nf: Int, optional
            Number of time points to evaluate.The default is 101,
        Ns: int, optional
            Number of samples of parameters to take. The default is 1000.
        nc: int, optional
        epsilon: np.float64, optional
            Acceptable error in leap. The default is 0.03.
        tau_update_frequency : int, optional
        verbose: bool, optional
            Verbosity of output. The default is False.
        Ti: float, optional
            Start time of integrator. The default is 0.
        method: str, optional
            Pyross integrator to use. The default is "deterministic".
        events: list of python functions, optional
            List of events that the current state can satisfy to change behaviour
            of the contact matrix. Event occurs when the value of the function
            changes sign. Event.direction determines which direction triggers
            the event, takign values {+1,-1}.
            The default is [].
        contactMatricies: list of python functions
            New contact matrix after the corresponding event occurs
            The default is [].
        events_repeat: bool, optional
            Wheither events is periodic in time. The default is false.
        events_subsequent : bool, optional
            TODO
             
        Returns
        -------
        out_dict : dict
            Dictionary containing the following keys:
                X: list
                    List of resultant trajectories
                t: list
                    List of times at which X is evaluated.
                X_mean : list
                    Mean trajectory of X
                X_std : list
                    Standard devation of trajectories of X at each time point.
                <init params> : Initial parameters passed at object instantiation.
                sample_parameters : list of parameters sampled to make trajectories.
                sample_inits : List of initial state vectors tried.
        """

        cdef:
            int M=self.M, k_tot = self.k_tot
            double [:] means_params=self.means_params, means_init=self.means_init
            double [:,:] cov_params=self.cov_params, cov_init=self.cov_init
            double start_time, end_time
            np.ndarray trajectories = np.zeros( [Ns,k_tot*M,Nf] , dtype=DTYPE)
            np.ndarray mean_traj = np.zeros([ k_tot*M, Nf] ,dtype=DTYPE)
            np.ndarray std_traj = np.zeros([ k_tot*M, Nf] ,dtype=DTYPE)
            np.ndarray sample_parameters, sample_inits
            bint control = False

        if (contactMatrix == None) and (len(events) == 0):
            raise RuntimeError("Neither contactMatrix function nor list of events provided.")

        if len(events) > 0:
            control = True

        sample_parameters = np.random.multivariate_normal(means_params, cov_params, Ns)
        sample_inits = np.random.multivariate_normal(means_init, cov_init, Ns)

        start_time = timer()
        for i in range(Ns):
            if verbose:
                print('Running simulation {0} of {1}'.format(i+1,Ns),end='\r')
            while (sample_parameters[i] < 0).any():
                sample_parameters[i] = np.random.multivariate_normal(means_params,cov_params)
            while (sample_inits[i] < 0).any():
                sample_inits[i] = np.random.multivariate_normal(means_init, cov_init)
            parameters = {'alpha':sample_parameters[i,0],
                        'beta':sample_parameters[i,1],
                        'gIa':sample_parameters[i,2],
                        'gIs':sample_parameters[i,3],
                        'gE':sample_parameters[i,4],
                        'gA':sample_parameters[i,5],
                        'gIh':self.gIh,'gIc':self.gIc,
                        'fsa':self.fsa,'fh':self.fh,
                        'sa':self.sa,'hh':self.hh,
                        'mm':self.mm,'cc':self.cc
                          }
            #
            S0 = ( sample_inits[i, : M] * self.N).astype('int')
            E0 = (sample_inits[i, M: 2*M]* self.N).astype('int')
            A0 = (sample_inits[i, 2*M: 3*M]* self.N).astype('int')
            Ia0 = (sample_inits[i, 3*M: 4*M]* self.N).astype('int')
            Is0 = (sample_inits[i, 4*M: 5*M]* self.N).astype('int')
            Ih0 = (sample_inits[i, 5*M: 6*M]* self.N).astype('int')
            Ic0 = (sample_inits[i, 6*M: 7*M]* self.N).astype('int')
            Im0 = (sample_inits[i, 7*M: 8*M]* self.N).astype('int')
            #
            if control:
                model = pyross.control.SEAI5R(parameters, M, self.Ni)
                cur_result = model.simulate(S0, E0, A0, Ia0, Is0, Ih0, Ic0, Im0,
                              events, contactMatrices, Tf, Nf,
                              events_repeat=events_repeat,
                              events_subsequent=events_subsequent,
                              nc=nc,epsilon=epsilon,
                              tau_update_frequency=tau_update_frequency,
                              method=method)
            else:
                if method == 'deterministic':
                    model = pyross.deterministic.SEAI5R(parameters, M, self.Ni)
                    cur_result = model.simulate(S0, E0, A0, Ia0, Is0, Ih0, Ic0, Im0,
                                    contactMatrix,
                                    Tf, Nf)
                else:
                    model = pyross.stochastic.SEAI5R(parameters, M, self.Ni)
                    cur_result = model.simulate(S0, E0, A0, Ia0, Is0, Ih0, Ic0, Im0,
                                  contactMatrix, Tf, Nf,
                                  nc=nc,epsilon=epsilon,
                                  tau_update_frequency=tau_update_frequency,
                                  method=method)
            #
            trajectories[i] = (cur_result['X'][:, :k_tot*M]).T
        end_time = timer()
        if verbose:
            print('Finished. Time needed for evaluation:',time.strftime('%H:%M:%S',time.gmtime( end_time-start_time)) )
        #

        mean_traj = np.mean(trajectories,axis=0)
        std_traj = np.sqrt( np.var(trajectories,axis=0) )

        out_dict = {'X':trajectories, 't':cur_result['t'],
                    'X_mean':mean_traj,'X_std':std_traj,
                      'Ni':self.Ni, 'M':self.M,
                      'alpha':self.alpha, 'beta':self.beta,
                      'gIa':self.gIa,'gIs':self.gIs,
                      'gIh':self.gIh,'gIc':self.gIc,
                      'fsa':self.fsa,'fh':self.fh,
                      'gE':self.gE,'gA':self.gA,
                      'sa':self.sa,'hh':self.hh,
                      'mm':self.mm,'cc':self.cc,
                      'cov_params':self.cov_params,
                      'cov_init':self.cov_init,
                      'sample_parameters':sample_parameters,
                      'sample_inits':sample_inits
                      }
        return out_dict



# distutils: language = c++
# distutils: extra_compile_args = -std=c++11

import numpy as np
cimport numpy as np
cimport cpython
#from cython.parallel import prange
DTYPE   = np.float
ctypedef np.float_t DTYPE_t

import pyross.stochastic as stochastic
import pyross.deterministic as deterministic

cdef class SIR:
    """
    Susceptible, Infected, Removed (SIR)
    Ia: asymptomatic
    Is: symptomatic

    ...

    Parameters
    ----------
    parameters: dict
        Contains the following keys:
            alpha: float, np.array (M,)
                fraction of infected who are asymptomatic.
            beta: float
                rate of spread of infection.
            gIa: float
                rate of removal from asymptomatic individuals.
            gIs: float
                rate of removal from symptomatic individuals.
            fsa: float
                fraction by which symptomatic individuals do not self-isolate.
    M: int
        Number of compartments of individual for each class.
        I.e len(contactMatrix)
    Ni: np.array(3*M, )
        Initial number in each compartment and class

    Methods
    -------
    simulate
    """
    cdef:
        readonly int N, M,
        readonly double alpha, beta, gIa, gIs, fsa
        readonly np.ndarray Ni
        readonly dict parameters

    def __init__(self, parameters, M, Ni):

        self.M     = M
        self.Ni    = np.zeros( self.M, dtype=DTYPE)
        self.Ni    = Ni
        self.parameters = parameters

    cdef below_threshold(self,populations,thresholds):
        cdef:
            int M = self.M
        for i in range(3):
            for j in range(M):
                if (populations[i][j] < thresholds[i][j]):
                    return True
        return False

    cdef find_passing_index(self,trajectory,thresholds):
        cdef:
            int min_index = 0
            int length_of_traj = len(trajectory)
            int M = self.M
        for i in range(length_of_traj-1):
            for j in range(3):
                for k in range(M):
                    product = (trajectory[i,j*M + k] - thresholds[j][k])*(trajectory[i+1,j*M + k] - thresholds[j][k])
                    if product < 0:
                        return min_index+i+1 # index of first datapoint past threshold
        return min_index


    cpdef simulate(self, S0, Ia0, Is0, contactMatrix, long Tf,
                                dict thresholds,
                                double dt_stoch = 20,double dt_det = 300,
                                method='gillespie',
                                int nc = 30, double epsilon = 0.03,
                                int tau_update_frequency = 1,
                                ):
        cdef:
            int M=self.M
            tuple thresholds_from_below = thresholds.get('from_below')
            tuple thresholds_from_above = thresholds.get('from_above')
            tuple cur_populations = (S0, Ia0, Is0)
            int cur_t = 0
            long dt, cur_Tf, cur_Nt

        t_arr = np.arange(Tf+1,dtype=int)
        trajectory = np.zeros([Tf+1,3*M],
                              dtype=long)
        trajectory[0] = np.concatenate((S0,Ia0,Is0))

        # initialize both stochastic and deterministic simulations
        model_stoch = stochastic.SIR(self.parameters, M, self.Ni)
        model_det = deterministic.SIR(self.parameters, M, self.Ni)
        # check if initially we are above or below the threshold (from below)
        below = self.below_threshold(cur_populations,thresholds_from_below)

        while (cur_t < Tf):
            # run simulation
            if below:
                dt = long(np.round(dt_stoch))
            else:
                dt = long(np.round(dt_det))
            # duration of simulation
            cur_Tf = long( np.round( np.min([dt,(Tf-cur_t)]) ) )
            cur_Nt = long( np.round(cur_Tf+1 ) )
            #

            if below:
                cur_result = model_stoch.simulate(*cur_populations,
                                                     contactMatrix,
                                                     cur_Tf, cur_Nt,
                                                     method=method,epsilon=epsilon,
                                                     tau_update_frequency=tau_update_frequency,
                                                     )
                cur_traj = cur_result['X']
            else:
                cur_result = model_det.simulate(*cur_populations,
                                                     contactMatrix,
                                                     cur_Tf, cur_Nt)
                cur_traj = np.array(np.round(cur_result['X']),dtype=int)

            # check if we passed the threshold
            if below:
                passing_index = self.find_passing_index(cur_traj,thresholds_from_below)
            else:
                passing_index = self.find_passing_index(cur_traj,thresholds_from_above)
            if passing_index == 0: # means we have not passed through the threshold
                trajectory[cur_t+1:cur_t+cur_Nt] = cur_traj[1:]
                cur_t += cur_Tf
                cur_populations = ( cur_traj[-1,:M,],
                                   cur_traj[-1,M:2*M],
                                   cur_traj[-1,2*M:] )
            else: # means we passed through the threshold
                if below: # this means we came from below
                    below = False # now we are above
                else: # this means we came from above
                    below = True # now we are below
                trajectory[cur_t+1:cur_t+passing_index+1] = cur_traj[1:passing_index+1]
                # Note that the zero-th element of cur_traj is
                # the last datapoint from the previous simulation
                cur_t += passing_index
                cur_populations = ( cur_traj[passing_index,:M],
                                   cur_traj[passing_index,M:2*M],
                                   cur_traj[passing_index,2*M:] )
        out_dict = {'X':trajectory, 't':t_arr,
                     'Ni':self.Ni, 'M':self.M,
                     'alpha':self.alpha, 'beta':self.beta,
                     'gIa':self.gIa, 'gIs':self.gIs,
                      'from_below':thresholds_from_below,
                      'from_above':thresholds_from_above,
                      'dt_det':dt_det,'dt_stoch':dt_stoch}
        return out_dict



import  numpy as np
cimport numpy as np
cimport cython
from libc.math cimport sqrt, pow, log
from cython.parallel import prange
cdef double PI = 3.1415926535
from scipy.sparse import spdiags
import matplotlib.pyplot as plt




def get_IC(ep, M, Ni, Nc, Nk, Cij, T, Td, tsi, beta):
    #Initial susceptible is the whole population
    S_0 = Ni

    #get I_0 from linear stability analysis
    #find the fastest growing linear mode
    A = np.matmul(np.diag(Ni),Cij)
    A = np.matmul(A,np.diag(1/Ni))
    sp = np.linspace(0,T,1000)
    lam = np.log(2)/Td;
    A = A*np.trapz(np.exp(-lam*sp)*np.interp(sp,tsi,beta),sp)
    w, v = np.linalg.eig(-np.identity(M) + A)

    #now identify the largest eigenvalue/eigenvector...
    pos = np.where(w == np.amax(w))
    pos = pos[0][0]
    lam = T/Td*np.log(2)
    s = np.linspace(-1,1,Nk)
    I_0 = ep*np.abs(np.real(np.outer(np.exp(-lam*s),v[:,pos])))

    #just set Ic_0 to zero -- these numbers are too small to matter
    Ic_0 = np.zeros((Nc,M))
    return S_0, I_0, Ic_0



def age_dep_rates(rate, int M, str name):
    if np.size(rate)==1:
        return rate*np.ones(M)
    elif np.size(rate)==M:
        return rate
    else:
        raise Exception('{} can be a number or an array of size M'.format(name))




class GPR:
    def __init__(self, nS, nT, iP, nP, xS, xT, yT):
        self.nS   =  nS           # # of test data points
        self.nT   =  nT           # # of training data points
        self.iP   =  iP           # # inverse of sigma
        self.nP   =  nP           # # number of priors
        self.xS   =  xS           # test input
        self.xT   =  xT           # training input
        self.yT   =  yT           # training output

        self.yS   =  0            # test output
        self.yP   =  0            # prior output
        self.K    =  0            # kernel
        self.Ks   =  0            # kernel
        self.Kss  =  0            # kernel
        self.mu   =  0            # mean
        self.sd   =  0            # stanndard deviation


    def calcDistM(self, r, s):
        '''Calculate distance matrix between 2 1D arrays'''
        return r[..., np.newaxis] - s[np.newaxis, ...]


    def calcKernels(self):
        '''Calculate the kernel'''
        cc = self.iP*0.5
        self.K   = np.exp(-cc*self.calcDistM(self.xT, self.xT)**2)
        self.Ks  = np.exp(-cc*self.calcDistM(self.xT, self.xS)**2)
        self.Kss = np.exp(-cc*self.calcDistM(self.xS, self.xS)**2)
        return


    def calcPrior(self):
        '''Calculate the prior'''
        L  = np.linalg.cholesky(self.Kss + 1e-6*np.eye(self.nS))
        G  = np.random.normal(size=(self.nS, self.nP))
        yP = np.dot(L, G)
        return


    def calcMuSigma(self):
        '''Calculate the mean'''
        self.mu =  np.dot(self.Ks.T, np.linalg.solve(self.K, self.yT))

        vv = self.Kss - np.dot(self.Ks.T, np.linalg.solve(self.K, self.Ks))
        self.sd = np.sqrt(np.abs(np.diag(vv)))

        # Posterior
        L       = np.linalg.cholesky(vv + 1e-6*np.eye(self.nS))
        self.yS = self.mu.reshape(-1,1) + np.dot(L, np.random.normal(size=(self.nS, self.nP)))
        return


    def plotResults(self):
        plt.plot(self.xT, self.yT, 'o', ms=10, mfc='#348ABD', mec='none', label='training set' )
        plt.plot(self.xS, self.yS, '#dddddd', lw=1.5, label='posterior')
        plt.plot(self.xS, self.mu, '#A60628', lw=2, label='mean')

        # fill 95% confidence interval (2*sd about the mean)
        plt.fill_between(self.xS.flat, self.mu-2*self.sd, self.mu+2*self.sd, color="#348ABD", alpha=0.4, label='2 sigma')
        plt.axis('tight'); plt.legend(fontsize=15); plt.rcParams.update({'font.size':18})


    def runGPR(self):
        self.calcKernels()
        self.calcPrior()
        self.calcMuSigma()
        self.plotResults()


def parse_model_spec(model_spec, param_keys):

    # First, extract the classes
    class_list = model_spec['classes']
    nClass = len(class_list) # total number of classes

    # Make dictionaries for class and parameter index look-up
    class_index_dict = {class_name:i for (i, class_name) in enumerate(class_list)}
    params_index_dict = {param: i for (i, param) in enumerate(param_keys)}

    try:
        # dictionaries of all linear terms and all infection terms
        linear_dict = {class_name: model_spec[class_name]['linear'] for class_name in class_list }
        infection_dict = {class_name: model_spec[class_name]['infection'] for class_name in class_list }

        # parse the linear terms into a list of [rate_index, reagent_index] and a dictionary for the product
        linear_terms_set = set() # used to check for duplicates
        linear_terms_list = [] # collect all linear terms
        linear_terms_destination_dict = {} # a dictionary for the product
        for (k, val) in linear_dict.items():
            for (reagent, rate) in val:
                if (reagent, rate) in linear_terms_set:
                    raise Exception('Duplicates linear terms: {}, {}'.format(reagent, rate))
                else:
                    linear_terms_set.add((reagent, rate))
                    reagent_index = class_index_dict[reagent]
                    if rate.startswith('-'):
                        rate = rate[1:]
                        rate_index = params_index_dict[rate]
                        linear_terms_list.append([rate_index, reagent_index, -1])
                    else:
                        rate_index = params_index_dict[rate]
                        linear_terms_destination_dict[(rate_index, reagent_index)] = class_index_dict[k]

        # parse the infection terms into a list of [rate_index, reagent_index] and a dictionary for the product
        infection_terms_set = set() # used to check to duplicates
        infection_terms_list = [] # collect all infection terms
        infection_terms_destination_dict = {} # a dictionary for the product
        for (k, val) in infection_dict.items():
            for (reagent, rate) in val:
                if (reagent, rate) in infection_terms_set:
                    raise Exception('Duplicates infection terms: {}, {}'.format(reagent, rate))
                else:
                    infection_terms_set.add((reagent, rate))
                    reagent_index = class_index_dict[reagent]
                    if rate.startswith('-'):
                        rate = rate[1:]
                        if k != 'S':
                            raise Exception('A susceptible group that is not S: {}'.format(k))
                        else:
                            rate_index = params_index_dict[rate]
                            infection_terms_list.append([rate_index, reagent_index, -1])
                    else:
                        rate_index = params_index_dict[rate]
                        infection_terms_destination_dict[(rate_index, reagent_index)] = class_index_dict[k]

    except KeyError:
        raise Exception('No reactions for some classses. Please check model_spec again')

    # set the product index
    set_destination(linear_terms_list, linear_terms_destination_dict)
    set_destination(infection_terms_list, infection_terms_destination_dict)
    return nClass, class_index_dict, np.array(linear_terms_list, dtype=np.intc), np.array(infection_terms_list, dtype=np.intc)

def set_destination(term_list, destination_dict):
    '''
    A function used by parse_model_spec that sets the product_index
    '''
    for term in term_list:
        rate_index = term[0]
        reagent_index = term[1]
        if (rate_index, reagent_index) in destination_dict.keys():
            product_index = destination_dict[(rate_index, reagent_index)]
            term[2] = product_index

def age_dep_rates(rate, int M, str name):
    if np.size(rate)==1:
        return rate*np.ones(M)
    elif np.size(rate)==M:
        return rate
    else:
        raise Exception('{} can be a number or an array of size M'.format(name))

def make_log_norm_dist(means, stds):
    var = stds**2
    means_sq = means**2
    scale = means_sq/np.sqrt(means_sq+var)
    s = np.sqrt(np.log(1+var/means_sq))
    return s, scale


DTYPE = np.float

cpdef forward_euler_integration(f, double [:] x, double t1, double t2, Py_ssize_t steps):
    cdef:
        double dt=(t2-t1)/(steps-1),t=t1
        double [:] fx
        Py_ssize_t i, j, size=x.shape[0]
        double [:, :] sol=np.empty((steps, size), dtype=DTYPE)

    for j in range(size):
        sol[0, j] = x[j]
    for i in range(1, steps):
        fx = f(t, x)
        for j in range(size):
            sol[i, j] = sol[i-1, j] + fx[j]*dt
        t += dt
    return sol

cpdef RK2_integration(f, double [:] x, double t1, double t2, Py_ssize_t steps):
    cdef:
        double dt=(t2-t1)/(steps-1),t=t1
        double [:] fx
        Py_ssize_t i, j, size=x.shape[0]
        double [:] k1=np.empty((size), dtype=DTYPE), temp=np.empty((size), dtype=DTYPE)
        double [:, :] sol=np.empty((steps, size), dtype=DTYPE)

    for j in range(size):
        sol[0, j] = x[j]
    for i in range(1, steps):
        fx = f(t, x)
        for j in range(size):
            k1[j] = dt*fx[j]
            temp[j] = sol[i-1, j] + k1[j]
        t += dt
        fx = f(t, temp)
        for j in range(size):
            sol[i, j] = sol[i-1, j] + 0.5*(k1[j] + dt*fx[j])
    return sol

def plotSIR(data, showPlot=True):
    t = data['t']
    X = data['X']
    M = data['M']
    Ni = data['Ni']
    N = Ni.sum()

    S = X[:, 0: M]
    Is = X[:, 2*M:3*M]
    R = Ni - X[:, 0:M] - X[:, M:2*M] - X[:, 2*M:3*M]


    sumS = S.sum(axis=1)
    sumI = Is.sum(axis=1)
    sumR = R.sum(axis=1)

    plt.fill_between(t, 0, sumS/N, color="#348ABD", alpha=0.3)
    plt.plot(t, sumS/N, '-', color="#348ABD", label='$S$', lw=4)


    plt.fill_between(t, 0, sumI/N, color='#A60628', alpha=0.3)
    plt.plot(t, sumI/N, '-', color='#A60628', label='$I$', lw=4)

    plt.fill_between(t, 0, sumR/N, color="dimgrey", alpha=0.3)
    plt.plot(t, sumR/N, '-', color="dimgrey", label='$R$', lw=4)

    plt.legend(fontsize=26); plt.grid()
    plt.autoscale(enable=True, axis='x', tight=True)

    plt.ylabel('Fraction of compartment value')
    plt.xlabel('Days')

    if True != showPlot:
        pass
    else:
        plt.show()



# This file implements two methods to compute the evidence of an epidemiological model. These are an addition to nested sampling 
# which can be used to estimate the evidence and which is already implemented in `pyross.inference`. However, for complicated, 
# high-dimensional models, nested sampling can be very slow. The methods implemented here are applicable whenever MCMC sampling 
# of the posterior is feasible and should work even for sophisticated epidemiological models. 

import numpy as np
try:
    # Support for MCMC sampling.
    import emcee
except ImportError:
    emcee = None

try:
    # Optional support for multiprocessing.
    import pathos.multiprocessing as pathos_mp
except ImportError:
    pathos_mp = None

import pyross.utils
from pyross.utils_python import Prior

DTYPE = np.float

def get_parameters(estimator, x, Tf, prior_dict, contactMatrix=None, generator=None, intervention_fun=None, 
                   tangent=False):
    """Process an estimator from `pyross.inference` to generate input arguments for the evidence computations 
    `pyross.evidence.evidence_smc` and `pyross.evidence.evidence_path_sampling` for estimation problems without latent 
    variables. The input has the same structure as the input of `pyross.inference.infer`, see there for a detailed 
    documentation of the arguments.

    Parameters
    ----------
    estimator: pyross.inference.SIR_type
        The estimator object of the underlying (non-latent) inference problem.
    x: 2d numpy.array
    Tf: float
    prior_dict: dict
    contactMatrix: callable, optional
    generator: pyross.contactMatrix, optional
    intervention_fun: callable, optional
    tangent: bool, optional

    Returns
    -------
    logl:
        The log-likelihood of the inference problem.
    prior:
        The prior distribution of the parameters.
    ndim:
        The number of (flat) parameters.
    """
    # Sanity checks of the inputs
    if (contactMatrix is None) == (generator is None):
        raise Exception('Specify either a fixed contactMatrix or a generator')
    if (intervention_fun is not None) and (generator is None):
        raise Exception('Specify a generator')
    if contactMatrix is not None:
        estimator.set_contact_matrix(contactMatrix)

    # Read in parameter priors
    prior_names, keys, guess, stds, bounds, \
    flat_guess_range, is_scale_parameter, scaled_guesses  \
            = pyross.utils.parse_param_prior_dict(prior_dict, estimator.M)
    prior = pyross.Prior(prior_names, bounds, guess, stds)

    logl = lambda params: estimator._loglikelihood(params, contactMatrix=contactMatrix, generator=generator, 
                intervention_fun=intervention_fun, keys=keys, x=x, Tf=Tf, tangent=tangent,
                is_scale_parameter=is_scale_parameter, flat_guess_range=flat_guess_range,
                scaled_guesses=scaled_guesses, bounds=bounds)
    
    return logl, prior, len(guess)


def latent_get_parameters(estimator, obs, fltr, Tf, param_priors, init_priors, contactMatrix=None, generator=None, 
                          intervention_fun=None, tangent=False, smooth_penalty=False, disable_bounds=False):
    """Process an estimator from `pyross.inference` to generate input arguments for the evidence computations 
    `pyross.evidence.evidence_smc` and `pyross.evidence.evidence_path_sampling` for estimation problems with latent 
    variables. The input has the same structure as the input of `pyross.inference.latent_infer`, see there for a detailed 
    documentation of the arguments.

    Parameters
    ----------
    estimator: pyross.inference.SIR_type
        The estimator object of the underlying (non-latent) inference problem.
    obs:  np.array
    fltr: 2d np.array
    Tf: float
    param_priors: dict
    init_priors: dict
    contactMatrix: callable, optional
    generator: pyross.contactMatrix, optional
    intervention_fun: callable, optional
    tangent: bool, optional

    Returns
    -------
    logl:
        The log-likelihood of the inference problem.
    prior:
        The prior distribution of the parameters.
    ndim:
        The number of (flat) parameters.
    """
    # Sanity checks of the intputs
    estimator._process_contact_matrix(contactMatrix, generator, intervention_fun)

    # Process fltr and obs
    fltr, obs, obs0 = pyross.utils.process_latent_data(fltr, obs)

    # Read in parameter priors
    param_prior_names, keys, param_guess, param_stds, param_bounds, param_guess_range, \
    is_scale_parameter, scaled_param_guesses \
        = pyross.utils.parse_param_prior_dict(param_priors, estimator.M)

    # Read in initial conditions priors
    init_prior_names, init_guess, init_stds, init_bounds, init_flags, init_fltrs \
        = pyross.utils.parse_init_prior_dict(init_priors, estimator.dim, len(obs0))

    # Concatenate the flattend parameter guess with init guess
    param_length = param_guess.shape[0]
    guess = np.concatenate([param_guess, init_guess]).astype(DTYPE)
    stds = np.concatenate([param_stds,init_stds]).astype(DTYPE)
    bounds = np.concatenate([param_bounds, init_bounds], axis=0).astype(DTYPE)

    prior = Prior(param_prior_names+init_prior_names, bounds, guess, stds)
    
    if not disable_bounds:
        logl = lambda params: estimator._loglikelihood_latent(params, generator=generator, intervention_fun=intervention_fun, 
                    param_keys=keys, param_guess_range=param_guess_range, is_scale_parameter=is_scale_parameter, 
                    scaled_param_guesses=scaled_param_guesses, param_length=param_length, obs=obs, fltr=fltr, Tf=Tf, obs0=obs0,
                    init_flags=init_flags, init_fltrs=init_fltrs, tangent=tangent, smooth_penalty=smooth_penalty, bounds=bounds)
    
    else:
        logl = lambda params: estimator._loglikelihood_latent(params, generator=generator, intervention_fun=intervention_fun, 
                    param_keys=keys, param_guess_range=param_guess_range, is_scale_parameter=is_scale_parameter, 
                    scaled_param_guesses=scaled_param_guesses, param_length=param_length, obs=obs, fltr=fltr, Tf=Tf, obs0=obs0,
                    init_flags=init_flags, init_fltrs=init_fltrs, tangent=tangent, smooth_penalty=smooth_penalty, bounds=None)

    return logl, prior, len(guess)


def compute_ess(weights):
    """Compute the effective sample size of a weighted set of samples."""
    w = weights.copy() / sum(weights)
    return 1/np.sum(w**2)


def compute_cess(old_weights, weights):
    """Compute the conditional effective sample size as decribed in [Zhou, Johansen, Aston 2016]."""
    return np.sum(old_weights * weights)**2 / np.sum(old_weights * weights**2)


def resample(N, particles, logl, probs):
    """ Implements the residual resampling scheme, see for example
    [Doucet, Johansen 2008], https://www.stats.ox.ac.uk/~doucet/doucet_johansen_tutorialPF2011.pdf
    """
    counts_det = np.floor(N * probs).astype('int')
    remaining_probs = probs - 1/N * counts_det
    remaining_probs /= np.sum(remaining_probs)
    counts_mult = np.random.multinomial(N - np.sum(counts_det), remaining_probs)
    counts = counts_det + counts_mult
    
    # Select the corresponding configurations
    ndim = particles.shape[1]
    result_particles = np.zeros((N, ndim))
    result_logl = np.zeros(N)
    index = 0
    for i in range(len(counts)):
        for _ in range(counts[i]):
            result_particles[index, :] = particles[i,:]
            result_logl[index] = logl[i]
            index += 1
        
    return result_particles, result_logl


def evidence_smc(logl, prior, ndim, npopulation=200, target_cess=0.9, min_ess=0.6, mcmc_iter=50, nprocesses=0, 
                 save_samples=True, verbose=True):
    """ Compute the evidence using an adaptive sequential Monte Carlo method.

    This function computes the model evidence of the inference problem using a sequential Monte Carlo particle method 
    starting at the prior distribution. We implement the method `SMC2` described in
    [Zhou, Johansen, Aston 2016], https://doi.org/10.1080/10618600.2015.1060885

    We start by sampling `npopulation` particles from the prior distribution with uniform weights. The target distribution of the weighted 
    set of particles gets transformed to the posterior distribution by a geometric annealing schedule. The step size is chosen adaptively
    based on the target decay rate of the effective samples size `target_cess` in each step. Once the effective sample size of the 
    weighted particles goes below `min_cess * npopulation`, we replace the weighted set of samples by a resampled, unweighted set. 
    Between each step, the particles are decorrelated and equilibreated on the current level distribution by running an MCMC chain.

    Parameters
    ----------
    logl:
        Input from `pyross.evidence.get_parameters` or `pyross.evidence.latent_get_parameters`.
    prior:
        Input from `pyross.evidence.get_parameters` or `pyross.evidence.latent_get_parameters`.
    ndim:
        Input from `pyross.evidence.get_parameters` or `pyross.evidence.latent_get_parameters`.
    npopulation: int
        The number of particles used for the SMC iteration. Higher number of particles increases the accuracy of the result.
    target_cess: float
        The target rate for the decay of the effective sample size (ess reduces by `1-target_cess` each step). Smaller
        values result in more iterations and a higher accuracy result.
    min_ess: float
        The minimal effective sample size of the system. Low effective sample size imply many particles in low probability
        regions. However, resampling adds variance so it should not be done in every step.
    mcmc_iter: int
        The number of MCMC iterations in each step of the algorithm. The number of iterations should be large enough to equilibrate
        the particles with the current distribution, higher iteration numbers will typically not result in more accurate results.
        Oftentimes, it makes more sense to increase the number of steps (via `target_cess`) instead of increasing the number of 
        iterations. This decreases the difference in distribution between consecutive steps and reduced the error of the final result. 
        This number should however be large enough to allow equal-position particles (that occur via resampling) to diverge from each 
        other.
    nprocesses: int
        The number of processes passed to the `emcee` MCMC sampler. By default, the number of physical cores is used. 
    save_samples: bool
        If true, this function returns the interal state of each MCMC iteration.
    verbose: bool
        If true, this function displays the progress of each MCMC iteration in addition to basic progress information.

    Returns
    -------
    log_evidence: float
        The estimate of the log evidence.        
    if save_samples=True:

        result_samples: list of (float, emcee.EnsembleSampler)
            The list of samplers and their corresponding step `alpha`.
    """
    if emcee is None:
        raise Exception("MCMC sampling needs optional dependency `emcee` which was not found.")

    if nprocesses == 0:
        if pathos_mp:
            # Optional dependecy for multiprocessing (pathos) is installed.
            nprocesses = pathos_mp.cpu_count()
        else:
            nprocesses = 1

    if nprocesses > 1 and pathos_mp is None:
        raise Exception("The Python package `pathos` is needed for multiprocessing.")

    if nprocesses > 1:
        mcmc_pool = pathos_mp.ProcessingPool(nprocesses)
        mcmc_map = mcmc_pool.map
    else:
        mcmc_pool = None
        mcmc_map = map

    logp = lambda params: np.sum(prior.logpdf(params))

    alpha = 0.0
    delta_alpha = 0.0001  # Initial step should be small enough to avoid float overflows.
    log_evidence = 0.0

    # If `save_samples==True`, we save the initial position of the particles and the result of
    # each `emcee` run in this list.
    result_samples = []

    # Sample initial points (truncated log-normal within boundaries)
    points = np.random.rand(npopulation, ndim)
    particles = prior.ppf(points)

    if save_samples:
        result_samples.append((alpha, particles))

    iters = 0
    weights = 1/npopulation * np.ones(npopulation)
    logl_particles = np.array(mcmc_map(logl, particles))
    while True:
        iters += 1
        
        # Find next delta_alpha value (based on target_cess):
        def comp_cess(da):
            log_incr_weights = da * logl_particles
            new_weights = log_incr_weights - np.mean(log_incr_weights)
            new_weights = weights * np.exp(new_weights)
            new_weights = new_weights / np.sum(new_weights)
            return compute_cess(weights, np.exp(log_incr_weights))
        
        da_upper = delta_alpha
        da_lower = 0.0
        # Increase da_upper until it is an upper bound
        while comp_cess(da_upper) > target_cess:
            da_lower = da_upper
            da_upper = 2*da_upper

        # Binary search for optimal delta_alpha
        delta_alpha = 1/2 * (da_upper + da_lower)
        while np.abs(comp_cess(delta_alpha) - target_cess) / target_cess > 1e-4:
            if comp_cess(delta_alpha) > target_cess:
                da_lower = delta_alpha
            else:
                da_upper = delta_alpha
            delta_alpha = 1/2 * (da_upper + da_lower)
        
        if alpha + delta_alpha > 1:
            delta_alpha = 1 - alpha
        alpha = alpha + delta_alpha
        
        # Compute reweighting from alpha -> alpha + delta_alpha:
        log_incr_weights = delta_alpha * logl_particles
        new_weights = log_incr_weights - np.mean(log_incr_weights)
        new_weights = weights * np.exp(new_weights)
        new_weights = new_weights / np.sum(new_weights)

        ess = compute_ess(new_weights)
        cess = compute_cess(weights, np.exp(log_incr_weights))
        if ess/npopulation < min_ess:
            # The effective samples size has become smaller than the minimal target. We resample the
            # the particles according to their weight to remove particles with low probability and
            # increase the population in regions with high probability.
            particles, logl_particles = resample(npopulation, particles, logl_particles, new_weights)
            log_incr_weights = delta_alpha * logl_particles
            new_weights = 1/npopulation * np.ones(npopulation)
            log_evidence += np.log(1/npopulation * np.sum(np.exp(log_incr_weights)))
        else:
            # The effective sample size is still acceptable, therefore we avoid resampling since it
            # introduces a small bias every time.
            log_evidence += np.log(np.sum(weights * np.exp(log_incr_weights)))
        print("Iteration {}: alpha = {}, ESS = {}, CESS = {}, log_evidence = {}".format(iters, alpha, ess, cess, log_evidence))
        
        if alpha == 1.0:
            # There is no need to run the chain for `alpha==1`, so we stop here.
            break

        # Allow mixing of the particles by running an MCMC chain that tagrets the current distribution. We
        # use `emcee` here to make use of the ensemble of particles. The SMC method is failry robust to
        # slow mixing here. 
        logpost = lambda param, alpha: logp(param) + alpha * logl(param)
        sampler = emcee.EnsembleSampler(npopulation, ndim, logpost, pool=mcmc_pool, kwargs={'alpha':alpha})
        emcee_state = emcee.State(particles, log_prob=list(map(logp, particles)) + alpha*logl_particles)
        sampler.run_mcmc(emcee_state, mcmc_iter, progress=verbose)

        if save_samples:
            result_samples.append((alpha, sampler))

        particles = sampler.get_last_sample().coords
        logl_particles = 1/alpha * (sampler.get_last_sample().log_prob - np.array(mcmc_map(logp, particles)))
        weights = new_weights

    if mcmc_pool is not None:
        mcmc_pool.close()
        mcmc_pool.join()
        mcmc_pool.clear()

    if save_samples:
        return log_evidence, result_samples
    
    return log_evidence


def evidence_path_sampling(logl, prior, ndim, steps, npopulation=100, mcmc_iter=1000, nprocesses=0,
                           initial_samples=10, verbose=True, extend_step_list=None, extend_sampler_list=None):
    """ Compute the evidence using path sampling (thermodynamic integration).

    This function computes posterior samples for the distributions

        p_s \propto prior * likelihood^s

    for 0<s≤1, s ∈ steps, using ensemble MCMC. The samples can be used to estimate the evidence via

        log_evidence = \int_0^1 \E_{p_s}[log_likelihood] ds

    which is know as path sampling or thermodynamic integration.

    This function starts with sampling `initial_samples * npopulation` samples from the (truncated log-normal) prior.
    Afterwards, it runs an ensemble MCMC chain with `npopulation` ensemble members for `mcmc_iter` iterations. To 
    minimise burn-in, the iteration is started with the last sample of the chain with the closest step `s` that
    has already been computed. To extend the results of this function with additional steps, provide the to-be-extended 
    result via the optional arguments `extend_step_list` and `extend_sampler_list`.

    This function only returns the step list and the corresponding samplers. To compute the evidence estimate,
    use `pyross.evidence.evidence_path_sampling_process_result`.

    Parameters
    ----------
    logl:
        Input from `pyross.evidence.get_parameters` or `pyross.evidence.latent_get_parameters`.
    prior:
        Input from `pyross.evidence.get_parameters` or `pyross.evidence.latent_get_parameters`.
    ndim:
        Input from `pyross.evidence.get_parameters` or `pyross.evidence.latent_get_parameters`.
    steps: list of float
        List of steps `s` for which the distribution `p_s` is explored using MCMC. Should be in ascending order
        and not include 0.
    npopulation: int
        The population size of the MCMC ensemble sampler (see documentation of `emcee` for details).
    mcmc_iters: int
        The number of iterations of the MCMC chain for each s ∈ steps.
    nprocesses: int
        The number of processes passed to the `emcee` MCMC sampler. By default, the number of physical cores is used.
    initial_samples: int
        Compute `initial_samples * npopulation` independent samples as the result for s = 0.
    verbose: bool
        If true, this function displays the progress of each MCMC iteration in addition to basic progress information.
    extend_step_list: list of float
        Extends the result of an earlier run of this function if this argument and `extend_sampler_list` are provided.
    extend_sampler_list: list of emcee.EnsembleSampler
        Extends the result of an earlier run of this function if this argument and `extend_step_list` are provided.

    Returns
    -------
    step_list: list of float
        The steps `s` for which `p_s` has been sampled from (including 0). Always in ascending order. 
    sampler_list: list
        The list of emcee.EnsembleSamplers (and an array of prior samples at 0). 
    """
    if emcee is None:
        raise Exception("MCMC sampling needs optional dependency `emcee` which was not found.")

    if nprocesses == 0:
        if pathos_mp:
            # Optional dependecy for multiprocessing (pathos) is installed.
            nprocesses = pathos_mp.cpu_count()
        else:
            nprocesses = 1

    if nprocesses > 1 and pathos_mp is None:
        raise Exception("The Python package `pathos` is needed for multiprocessing.")

    if nprocesses > 1:
        mcmc_pool = pathos_mp.ProcessingPool(nprocesses)
    else:
        mcmc_pool = None

    logp = lambda params: np.sum(prior.logpdf(params))

    # Sample initial points (truncated log-normal within boundaries)
    points = np.random.rand(npopulation, ndim)
    first_init_positions = prior.ppf(points)

    step_list = []
    sampler_list = []
    if extend_step_list is None:
        points = np.random.rand(initial_samples*npopulation, ndim)
        ext_init_positions = prior.ppf(points)

        step_list = [0.0]
        sampler_list = [ext_init_positions]
    else:
        step_list = extend_step_list
        sampler_list = extend_sampler_list

    for i, step in enumerate(steps):
        logpost = lambda param, step: logp(param) + step * logl(param)
        
        print("step: {} ({}/{})".format(step, i+1, len(steps)))

        # Find the closest init position:
        diff_to_step = np.abs(np.array(step_list) - step)
        pos = np.argmin(diff_to_step)

        if diff_to_step[pos] == 0.0 and step != 0.0:
            # There is already an MCMC chain for this step. In this situation, we extend the existing
            # chain by mcmc_iter iterations.
            sampler_list[pos].run_mcmc(None, mcmc_iter, progress=verbose)
        else:
            # Find a good starting point for the new MCMC chain. We use the last sample of the closest
            # chain.
            if pos == 0:
                init_positions = first_init_positions
            else:
                init_positions = sampler_list[pos].get_last_sample().coords
        
            sampler = emcee.EnsembleSampler(npopulation, ndim, logpost, pool=mcmc_pool, kwargs={'step':step})        
            sampler.run_mcmc(init_positions, mcmc_iter, progress=verbose)
            
            step_list.append(step)
            sampler_list.append(sampler)

    if mcmc_pool is not None:
        mcmc_pool.close()
        mcmc_pool.join()
        mcmc_pool.clear()
        
    # Sort results by step (needed if adding intermediate steps)
    sorted_results = sorted(zip(step_list, sampler_list), key=lambda val: val[0])
    step_list = [x[0] for x in sorted_results]
    sampler_list = [x[1] for x in sorted_results]
        
    return step_list, sampler_list


def evidence_path_sampling_process_result(logl, prior, ndim, step_list, sampler_list, 
                                          burn_in=0, nprocesses=0):
    """ Compute the evidence estimate for the result of `pyross.evidence.evidence_path_sampling`.

    Parameters
    ----------
    logl:
        Input from `pyross.evidence.get_parameters` or `pyross.evidence.latent_get_parameters`.
    prior:
        Input from `pyross.evidence.get_parameters` or `pyross.evidence.latent_get_parameters`.
    ndim:
        Input from `pyross.evidence.get_parameters` or `pyross.evidence.latent_get_parameters`.
    step_list: list of float
        Output of `pyross.evidence.evidence_path_sampling`. The steps `s` for which `p_s` has been 
        sampled from (including 0). Always in ascending order. 
    sampler_list: list
        Output of `pyross.evidence.evidence_path_sampling`. The list of emcee.EnsembleSamplers 
        (and an array of prior samples at 0). 
    burn_in: float or np.array
        The number of initial samples that are discarded before computing the Monte Carlo average.
    nprocesses: int
        The number of processes used to compute the prior likelihood. By default, the number of physical 
        cores is used.

    Returns
    -------
    log_evidence: float
        The estimate of the log evidence.
    vals: list of float
        The Monte Carlo average of the log-likelihood for each s s ∈ step_list.
    """
    if nprocesses == 0:
        if pathos_mp:
            # Optional dependecy for multiprocessing (pathos) is installed.
            nprocesses = pathos_mp.cpu_count()
        else:
            nprocesses = 1

    if nprocesses > 1 and pathos_mp is None:
        raise Exception("The Python package `pathos` is needed for multiprocessing.")

    if nprocesses > 1:
        mcmc_pool = pathos_mp.ProcessingPool(nprocesses)
        mcmc_map = mcmc_pool.map
    else:
        mcmc_pool = None
        mcmc_map = map

    logp = lambda params: np.sum(prior.logpdf(params))

    if np.size(burn_in) == 1:
        local_burn_in = burn_in * np.ones(len(step_list)-1, 'int')
    else:
        local_burn_in = burn_in

    vals = np.zeros(len(step_list))
    # step == 0 (special case because we sampled these positions directly from the prior):
    llike = list(mcmc_map(logl, sampler_list[0]))
    vals[0] = np.mean(llike)

    for i in range(1, len(step_list)):
        step = step_list[i]
        chain = sampler_list[i].get_chain(discard=local_burn_in[i-1], flat=True)
        log_probs = sampler_list[i].get_log_prob(discard=local_burn_in[i-1], flat=True)
        # Get likelihood values from log_probs
        llike = 1/step * (log_probs - list(map(logp, chain)))
        vals[i] = np.mean(llike)

    log_evidence = 0.0
    for i in range(1, len(step_list)):
        log_evidence += 1/2 * (step_list[i] - step_list[i-1]) * (vals[i] + vals[i-1])

    if mcmc_pool is not None:
        mcmc_pool.close()
        mcmc_pool.join()
        mcmc_pool.clear()

    return log_evidence, vals


def generate_traceplot(sampler, dims=None):
    """
    Generate a traceplot for an emcee.EnsembleSampler.

    Parameters
    ----------
    sampler: emcee.EnsembleSampler
        The sampler to plot the traceplot for.
    dims: list of int, optional
        Select the dimensions that are plotted. By default, all dimensions are selected.
    """
    if dims is None:
        dims = [i for i in range(sampler.ndim)]

    N = len(dims)
    fig, axes = plt.subplots(N, figsize=(12, N*10/8), sharex=True)
    samples = sampler.get_chain()
    for i in range(N):
        ax = axes[i]
        ax.plot(samples[:, :, dims[i]], "k", alpha=0.3)
        ax.set_xlim(0, len(samples))
    axes[-1].set_xlabel("step number")
